# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for transforms."""

from absl.testing import absltest
from absl.testing import parameterized
from dinosaur import coordinate_systems
from dinosaur import primitive_equations
from dinosaur import scales
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
import gin
import haiku as hk
import jax
import jax.numpy as jnp
from neuralgcm import transforms  # pylint: disable=unused-import
import numpy as np

NLAYER = 10
COORDS = coordinate_systems.CoordinateSystem(
    spherical_harmonic.Grid.T21(),
    sigma_coordinates.SigmaCoordinates.equidistant(NLAYER)
    )


class TransformsTest(parameterized.TestCase):
  """Tests transforms."""

  @parameterized.parameters(
      dict(),
  )
  def test_nondimensionalization_transforms(
      self,
  ):
    """Tests that mapping produces expected output structures."""
    gin.enter_interactive_mode()
    gin.clear_config()
    coords = coordinate_systems.CoordinateSystem(
        spherical_harmonic.Grid.T21(),
        sigma_coordinates.SigmaCoordinates.equidistant(6)
    )
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    dt = physics_specs.nondimensionalize(10 * scales.units.minute)
    aux_features = {}

    @gin.configurable
    def transform_forward(x, transform):  # defines a forward function
      return transform(coords, dt, physics_specs, aux_features, None)(x)

    units_map = {
        'divergence': '1 / second',
        'vorticity': '1 / second',
        'temperature_variation': 'kelvin',
        'log_surface_pressure': 'dimensionless',
        'tracers': {'specific_humidity': 'dimensionless', 'unused_val': 'm**2'},
    }
    matching_units_map = units_map | {'diagnostics': {}}
    matching_units_map['tracers'].pop('unused_val')
    config = '\n'.join([
        f'NondimensionalizeTransform.inputs_to_units_mapping = {units_map}',
        f'RedimensionalizeTransform.inputs_to_units_mapping = {units_map}',
        'nondim/transform_forward.transform = @NondimensionalizeTransform',
        'redim/transform_forward.transform = @RedimensionalizeTransform'
    ])
    gin.parse_config(config)

    inputs = {
        'diagnostics': {},
        'vorticity': jnp.ones(coords.nodal_shape),
        'divergence': jnp.ones(coords.nodal_shape),
        'temperature_variation': jnp.ones(coords.nodal_shape),
        'log_surface_pressure': jnp.ones(coords.surface_nodal_shape),
        'tracers': {
            'specific_humidity': jnp.ones(coords.nodal_shape)},
    }
    with self.subTest('nondimensionalization'):
      expected = jax.tree_util.tree_map(
          lambda x, y: physics_specs.nondimensionalize(x * scales.Quantity(y)),
          inputs, matching_units_map)
      with gin.config_scope('nondim'):
        transform_model = hk.without_apply_rng(hk.transform(transform_forward))
        actual = transform_model.apply(None, inputs)
        for x, y in zip(jax.tree_util.tree_leaves(expected),
                        jax.tree_util.tree_leaves(actual)):
          np.testing.assert_allclose(x, y)
    nondimensional_inputs = actual

    with self.subTest('redimensionalization'):
      expected = inputs
      with gin.config_scope('redim'):
        transform_model = hk.without_apply_rng(hk.transform(transform_forward))
        actual = transform_model.apply(None, nondimensional_inputs)
        for x, y in zip(jax.tree_util.tree_leaves(expected),
                        jax.tree_util.tree_leaves(actual)):
          np.testing.assert_allclose(x, y)

  def test_level_scale_transform(self):
    """Tests that level scaling works as expected."""
    gin.enter_interactive_mode()
    gin.clear_config()
    n_layers = 6
    coords = coordinate_systems.CoordinateSystem(
        spherical_harmonic.Grid.T21(),
        sigma_coordinates.SigmaCoordinates.equidistant(n_layers)
    )
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    dt = physics_specs.nondimensionalize(10 * scales.units.minute)
    aux_features = {}

    @gin.configurable
    def transform_forward(x, transform):  # defines a forward function
      return transform(coords, dt, physics_specs, aux_features)(x)

    level_scales = [1., 1., 1., 10., 100., 3.]
    config = '\n'.join([
        'scale/transform_forward.transform = @LevelScale',
        'inverse_scale/transform_forward.transform = @InverseLevelScale',
        f'LevelScale.scales = {level_scales}',
        'LevelScale.keys_to_scale = ("x", "z", "specific_humidity")',
        f'InverseLevelScale.scales = {level_scales}',
        'InverseLevelScale.keys_to_scale = ("x", "z", "specific_humidity")',
    ])
    gin.parse_config(config)

    vertical_values = np.arange(n_layers)[:, np.newaxis, np.newaxis]
    inputs = {
        'vorticity': np.ones(coords.nodal_shape) * vertical_values,
        'x': np.ones(coords.nodal_shape) * vertical_values,
        'tracers': {
            'specific_humidity': np.ones(coords.nodal_shape) * vertical_values,
            'y': np.ones(coords.nodal_shape) * vertical_values,
        },
    }
    transform_model = hk.without_apply_rng(hk.transform(transform_forward))

    with gin.config_scope('scale'):
      scaled = transform_model.apply(None, inputs)

    with self.subTest('where_scaling_is_applied'):
      expected_level_scale = np.asarray(level_scales)[:, np.newaxis, np.newaxis]
      np.testing.assert_allclose(
          scaled['tracers']['specific_humidity'],
          inputs['tracers']['specific_humidity'] * expected_level_scale)
      np.testing.assert_allclose(
          scaled['x'], inputs['x'] * expected_level_scale)
    with self.subTest('where_scaling_is_omitted'):
      np.testing.assert_allclose(scaled['vorticity'], inputs['vorticity'])
      np.testing.assert_allclose(scaled['tracers']['y'], inputs['tracers']['y'])

    with self.subTest('level_scale_round_trip'):
      with gin.config_scope('inverse_scale'):
        round_trip = transform_model.apply(None, scaled)
      jax.tree_util.tree_map(np.testing.assert_allclose, inputs, round_trip)

  @parameterized.named_parameters(
      dict(
          testcase_name='test_1_specific_humidity',
          inputs={
              'vorticity': np.ones(COORDS.nodal_shape),
              'divergence': np.ones(COORDS.nodal_shape),
              'temperature_variation': np.ones(COORDS.nodal_shape),
              'log_surface_pressure': np.ones(COORDS.surface_nodal_shape),
              'tracers': {
                  'specific_humidity': np.broadcast_to(
                      np.arange(NLAYER)[:, np.newaxis, np.newaxis],
                      (NLAYER,) + COORDS.horizontal.nodal_shape,
                  )
              },
          },
          expected_outputs={
              'vorticity': np.ones(COORDS.nodal_shape),
              'divergence': np.ones(COORDS.nodal_shape),
              'temperature_variation': np.ones(COORDS.nodal_shape),
              'log_surface_pressure': np.ones(COORDS.surface_nodal_shape),
              'tracers': {
                  'specific_humidity': np.broadcast_to(
                      np.arange(NLAYER)[2:8, np.newaxis, np.newaxis],
                      (6,) + COORDS.horizontal.nodal_shape,
                  )
              },
          },
          sigma_ranges={'specific_humidity': (0.2, 0.8)},
      ),
      dict(
          testcase_name='test_2_fields',
          inputs={
              'vorticity': np.ones(COORDS.nodal_shape),
              'divergence': np.ones(COORDS.nodal_shape),
              'temperature_variation': np.ones(COORDS.nodal_shape),
              'log_surface_pressure': np.ones(COORDS.surface_nodal_shape),
              'tracers': {
                  'specific_humidity': np.broadcast_to(
                      np.arange(NLAYER)[:, np.newaxis, np.newaxis],
                      (NLAYER,) + COORDS.horizontal.nodal_shape,
                  )
              },
          },
          expected_outputs={
              'vorticity': np.ones(COORDS.nodal_shape)[2:8, :, :],
              'divergence': np.ones(COORDS.nodal_shape),
              'temperature_variation': np.ones(COORDS.nodal_shape),
              'log_surface_pressure': np.ones(COORDS.surface_nodal_shape),
              'tracers': {
                  'specific_humidity': np.broadcast_to(
                      np.arange(NLAYER)[2:8, np.newaxis, np.newaxis],
                      (6,) + COORDS.horizontal.nodal_shape,
                  )
              },
          },
          sigma_ranges={
              'specific_humidity': (0.2, 0.8),
              'vorticity': (0.2, 0.8),
          },
      ),
  )
  def test_truncate_sigma_levels_transform(
      self, inputs, expected_outputs, sigma_ranges):
    """Tests TruncateSigmaLevels."""
    coords = COORDS
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    dt = physics_specs.nondimensionalize(10 * scales.units.minute)
    aux_features = {}

    def truncate_transform_fwd(x):
      return transforms.TruncateSigmaLevels(
          coords, dt, physics_specs, aux_features, sigma_ranges=sigma_ranges)(x)
    transform_model = hk.without_apply_rng(hk.transform(truncate_transform_fwd))
    actual_outputs = transform_model.apply(None, inputs)
    with self.subTest('dict'):
      actual_outputs = transform_model.apply(None, inputs)
      jax.tree_util.tree_map(
          np.testing.assert_allclose, expected_outputs, actual_outputs
      )
    with self.subTest('state'):
      state_inputs = primitive_equations.State(**inputs)
      expected_state_outputs = primitive_equations.State(**expected_outputs)
      actual_state_outputs = transform_model.apply(None, state_inputs)
      jax.tree_util.tree_map(
          np.testing.assert_allclose,
          expected_state_outputs,
          actual_state_outputs,
      )

  def test_take_surface_adjacent_sigma_level(self):
    """Tests TakeSurfaceAdjacentSigmaLevel."""
    coords = COORDS
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    dt = physics_specs.nondimensionalize(10 * scales.units.minute)
    aux_features = {}

    # define data that has value 1 only on the index=-1 sigma level
    volume_data = np.zeros(COORDS.nodal_shape)
    volume_data[-1] += 1
    surface_data = np.ones(COORDS.surface_nodal_shape)
    inputs = {
        'vorticity': volume_data,
        'divergence': volume_data,
        'temperature_variation': volume_data,
        'log_surface_pressure': surface_data,
        'tracers': {'specific_humidity': volume_data},
    }
    # expect sliced data to be shape (1, lon, lat) and have value 1
    expected_outputs = {
        'vorticity': surface_data,
        'divergence': surface_data,
        'temperature_variation': surface_data,
        'log_surface_pressure': surface_data,
        'tracers': {'specific_humidity': surface_data},
    }

    def transform_fwd(x):
      return transforms.TakeSurfaceAdjacentSigmaLevel(
          coords, dt, physics_specs, aux_features)(x)
    transform_model = hk.without_apply_rng(hk.transform(transform_fwd))
    actual_outputs = transform_model.apply(None, inputs)

    with self.subTest('dict'):
      actual_outputs = transform_model.apply(None, inputs)
      jax.tree_util.tree_map(
          np.testing.assert_allclose, expected_outputs, actual_outputs
      )

    with self.subTest('state'):
      state_inputs = primitive_equations.State(**inputs)
      expected_state_outputs = primitive_equations.State(**expected_outputs)
      actual_state_outputs = transform_model.apply(None, state_inputs)
      jax.tree_util.tree_map(
          np.testing.assert_allclose,
          expected_state_outputs,
          actual_state_outputs,
      )

  @parameterized.parameters(
      dict(transform_module=transforms.SoftClip, max_value=10.0),
      dict(transform_module=transforms.HardClip, max_value=15.5),
      dict(transform_module=transforms.SoftClip, max_value=50.0),
  )
  def test_clip_transforms(self, transform_module, max_value):
    """Tests that clipping transforms work as expected."""
    def transform_forward(x):  # defines a forward function
      return transform_module(None, None, None, None, max_value=max_value)(x)

    inputs = {
        'positive': np.arange(20),
        'negative': -np.arange(30),
        'both': np.arange(-30, 30),
        '~same': np.linspace(-max_value / 4, max_value / 4, 50),  # ~same.
        'tracers': {
            'c': np.ones(3),
            'y': np.arange(-30, 30),
        },
    }
    transform_model = hk.without_apply_rng(hk.transform(transform_forward))
    clipped = transform_model.apply(None, inputs)

    with self.subTest('valid_range'):
      jax.tree_util.tree_map(
          lambda x: self.assertGreaterEqual(max_value, x.max()), clipped)
      jax.tree_util.tree_map(
          lambda x: self.assertGreaterEqual(x.min(), -max_value), clipped)
    with self.subTest('near_identity_in_range'):
      jax.tree_util.tree_map(
          lambda x, y: np.testing.assert_allclose(x, y, atol=1e-3),
          inputs['~same'], clipped['~same'])

  @parameterized.parameters(
      dict(
          input_keys=('a', 'b'),
          regex_patterns='a',
          expected_out_keys=('a',)
      ),
      dict(
          input_keys=('a_1', 'b', 'a_2', 'a_3', 'b_1', 'ab_1', 't_a_1'),
          regex_patterns='a_.*',
          expected_out_keys=('a_1', 'a_2', 'a_3'),
      ),
      dict(
          input_keys=('a_1', 'b', 'a_2', 'a_3', 'b_1', 'ab_1', 't_a_1'),
          regex_patterns='a_.|ab.*',
          expected_out_keys=('a_1', 'a_2', 'a_3', 'ab_1'),
      ),
      dict(
          input_keys=('q1rpop', 'e1g', 'y2q', 'a_1e', 'b21t'),
          regex_patterns='.1.*',
          expected_out_keys=('q1rpop', 'e1g'),
      ),
  )
  def test_feature_selector(
      self,
      input_keys,
      regex_patterns,
      expected_out_keys,
  ):
    """Tests that FeatureSelector work as expected."""
    selector_fn = transforms.FeatureSelector(regex_patterns)
    values = np.arange(20)  # values to grab for inputs.
    inputs = {k: values[i] for i, k in enumerate(input_keys)}
    actual = selector_fn(inputs)
    expected = {
        k: values[i]
        for i, k in enumerate(input_keys) if k in expected_out_keys}
    self.assertDictEqual(expected, actual)

  def test_sigma_squash_transform(self):
    """Tests that SquashLevelsTransform preserves shape."""
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    transform_fn = transforms.SquashLevelsTransform(
        COORDS, 0.1, physics_specs, {},
        low_cutoffs=(0.05, 0.1),
        # Make the high cutoff > 1 so ground level variables will not be zero'd.
        high_cutoffs=(0.86, 1.1),
    )
    # COORDS.nodal_shape ~ (levels, lon, lat)
    values = np.ones(COORDS.nodal_shape)  # values to grab for inputs.
    inputs = {k: values for k in ('a', 'b', 'c', 'd')}
    inputs['sim_time'] = np.ones(shape=())
    inputs['tracers'] = {'sh': values}

    # log_surface_pressure is defined only at ground level.
    inputs['log_surface_pressure'] = values[:1]  # (1, lon, lat)

    # Our high_cutoffs ensured we do not squash at ground level.
    self.assertGreater(inputs['log_surface_pressure'].min(), 0.5)

    out = transform_fn(inputs)

    for x, y in zip(jax.tree_util.tree_leaves(inputs),
                    jax.tree_util.tree_leaves(out)):
      self.assertEqual(x.shape, y.shape)


if __name__ == '__main__':
  jax.config.parse_flags_with_absl()
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for stochastic."""

from absl.testing import absltest
from absl.testing import parameterized
from dinosaur import coordinate_systems
from dinosaur import primitive_equations
from dinosaur import scales
from dinosaur import sigma_coordinates
from dinosaur import spherical_harmonic
import haiku as hk
import jax
from neuralgcm import stochastic
import numpy as np
import tensorflow_probability.substrates.jax as tfp


tree_map = jax.tree_util.tree_map

# Effectively constant correlation time/length.
CONSTANT_CORRELATION_TIME_HRS = 24 * 365 * 1000  # 1000 years in hours
CONSTANT_CORRELATION_LENGTH_KM = 40_075 * 10  # 10x circumference of earth in km


@absltest.skipThisClass('Base class')
class BaseRandomFieldTest(parameterized.TestCase):
  """Base class for random field tests."""

  def setUp(self):
    super().setUp()
    self.physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    self.dt = self.physics_specs.nondimensionalize(1 * scales.units.hour)

  def check_correlation_length(
      self,
      nodal_samples,
      expected_correlation_length,
      coords,
  ):
    """Checks the correlation length of random field."""
    unused_n_samples, n_lngs, n_lats = nodal_samples.shape
    expected_corr_frac = expected_correlation_length / (
        2 * np.pi * coords.horizontal.radius
    )
    acorr_lat = tfp.stats.auto_correlation(
        # Mean autocorrelation in the lat direction at the longitude=0 line.
        nodal_samples[:, 0, :],
        axis=-1,
    ).mean(axis=0)
    # There are 2 * n_lats points in the circumference.
    fractional_corr_len_lat = np.argmax(acorr_lat < 0) / (2 * n_lats)
    self.assertBetween(
        fractional_corr_len_lat,
        expected_corr_frac * 0.5,
        expected_corr_frac * 3,
    )
    acorr_lng = tfp.stats.auto_correlation(
        # Mean autocorrelation in the lng direction at the latitude=0 line.
        nodal_samples[:, :, n_lats // 2],
        axis=-1,
    ).mean(axis=0)
    fractional_corr_len_lng = np.argmax(acorr_lng < 0) / n_lngs
    self.assertBetween(
        fractional_corr_len_lng,
        expected_corr_frac * 0.2,
        expected_corr_frac * 3,
    )

  def check_mean(
      self,
      nodal_samples,
      coords,
      expected_mean,
      variance,
      correlation_length,
      mean_tol_in_standard_errs,
  ):
    """Checks the mean (at every point & average) of nodal_samples."""
    n_samples, unused_n_lngs, unused_n_lats = nodal_samples.shape

    # Pointwise mean should be with tol with high probability everywhere.
    # Since we're testing many (lat/lon) points, we allow for some deviations.
    standard_error = np.sqrt(variance) / np.sqrt(n_samples) if variance else 0.0
    np.testing.assert_allclose(
        # 95% of points are within specified tol. There may be outliers.
        np.percentile(np.mean(nodal_samples, axis=0), 95),
        expected_mean,
        atol=mean_tol_in_standard_errs * standard_error,
    )
    np.testing.assert_allclose(
        # 100% of points are within a looser tol.
        np.mean(nodal_samples, axis=0),
        expected_mean,
        atol=2 * mean_tol_in_standard_errs * standard_error,
    )

    # Check average mean over whole earth (standard_error will be lower so this
    # is a good second check).
    expected_corr_frac = correlation_length / coords.horizontal.radius
    n_equivalent_integrated_samples = n_samples / expected_corr_frac**2
    if variance:
      standard_error = np.sqrt(variance) / np.sqrt(
          n_equivalent_integrated_samples
      )
    else:
      standard_error = 0.0
    np.testing.assert_allclose(
        np.mean(nodal_samples),
        expected_mean,
        atol=mean_tol_in_standard_errs * standard_error,
    )

  def check_variance(
      self,
      nodal_samples,
      coords,
      correlation_length,
      expected_variance,
      var_tol_in_standard_errs,
  ):
    expected_variance = expected_variance or 0.0
    expected_integrated_variance = (
        expected_variance * 4 * np.pi * coords.horizontal.radius**2
    )

    n_samples, unused_n_lngs, unused_n_lats = nodal_samples.shape
    # Integrating over the sphere we get additional statistical power since
    # points decorrelate.
    expected_corr_frac = correlation_length / coords.horizontal.radius
    n_equivalent_integrated_samples = n_samples / expected_corr_frac**2
    standard_error = np.sqrt(
        # The variance of a (normal) variance estimate is 2 σ⁴ / (n - 1).
        2
        * expected_integrated_variance**2
        / n_equivalent_integrated_samples
    )

    np.testing.assert_allclose(
        coords.horizontal.integrate(np.var(nodal_samples, axis=0)),
        expected_integrated_variance,
        atol=var_tol_in_standard_errs * standard_error,
        rtol=0.0,
    )

  def check_unconditional_and_trajectory_stats(
      self,
      coords,
      random_field,
      mean,
      variance,
      grid,
      correlation_length,
      correlation_time,
      run_mean_check=True,
      run_variance_check=True,
      run_correlation_length_check=True,
      run_correlation_time_check=True,
      mean_tol_in_standard_errs=4,
      var_tol_in_standard_errs=4,
  ):
    del grid  # unused.
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    dt = physics_specs.nondimensionalize(1 * scales.units.hour)

    # generating multiple trajectories of random fields.
    n_samples = 500
    unroll_length = 40
    init_rngs = jax.random.split(jax.random.PRNGKey(5), n_samples)
    initial_states = jax.vmap(random_field.unconditional_sample)(init_rngs)
    initial_values = jax.vmap(random_field.to_nodal_values)(initial_states.core)

    with self.subTest('to_modal'):
      modal_initial_values = jax.vmap(random_field.to_modal_values)(
          initial_states.core
      )
      to_nodal = random_field.coords.horizontal.to_nodal
      to_modal = random_field.coords.horizontal.to_modal
      np.testing.assert_allclose(
          to_nodal(to_modal(initial_values)),
          to_nodal(modal_initial_values),
          rtol=1e-3,
          atol=1e-3,
      )

    with self.subTest('unconditional_sample_shape'):
      self.assertEqual(
          initial_values.shape, (n_samples,) + coords.horizontal.nodal_shape
      )

    n_lats = len(coords.horizontal.latitudes)
    n_lngs = len(coords.horizontal.longitudes)
    self.assertTupleEqual((n_samples, n_lngs, n_lats), initial_values.shape)

    if run_correlation_length_check and variance is not None:
      with self.subTest('unconditional_sample_correlation_len'):
        self.check_correlation_length(
            initial_values, correlation_length, coords
        )

    if run_mean_check:
      with self.subTest('unconditional_sample_pointwise_mean'):
        self.check_mean(
            initial_values,
            coords,
            expected_mean=mean,
            variance=variance,
            correlation_length=correlation_length,
            mean_tol_in_standard_errs=mean_tol_in_standard_errs,
        )

    if run_variance_check:
      with self.subTest('unconditional_sample_integrated_var'):
        self.check_variance(
            initial_values,
            coords,
            correlation_length=correlation_length,
            expected_variance=variance,
            var_tol_in_standard_errs=var_tol_in_standard_errs,
        )

    def step_fn(c, _):
      next_c = jax.vmap(random_field.advance)(c)
      next_output = jax.vmap(random_field.to_nodal_values)(next_c.core)
      return (next_c, next_output)

    _, field_trajectory = jax.lax.scan(
        step_fn, initial_states, xs=None, length=unroll_length
    )
    field_trajectory = jax.device_get(field_trajectory)

    if run_correlation_time_check and variance is not None:
      with self.subTest('trajectory_correlation_time'):
        # Mean autocorrelation at the lat=lng=0 point.
        acorr = tfp.stats.auto_correlation(
            field_trajectory[:, :, n_lngs // 2, 0], axis=0
        ).mean(axis=1)
        sample_decorr_time = dt * np.argmax(acorr < 0)
        self.assertBetween(
            sample_decorr_time, correlation_time / 2, correlation_time * 2
        )

    final_sample = field_trajectory[-1]

    if run_correlation_length_check and variance is not None:
      with self.subTest('final_sample_correlation_len'):
        self.check_correlation_length(final_sample, correlation_length, coords)

    if run_mean_check:
      with self.subTest('final_sample_pointwise_mean'):
        self.check_mean(
            final_sample,
            coords,
            expected_mean=mean,
            variance=variance,
            correlation_length=correlation_length,
            mean_tol_in_standard_errs=mean_tol_in_standard_errs,
        )

    if run_variance_check:
      with self.subTest('final_sample_integrated_var'):
        self.check_variance(
            final_sample,
            coords,
            correlation_length=correlation_length,
            expected_variance=variance,
            var_tol_in_standard_errs=var_tol_in_standard_errs,
        )

  def check_independent(self, x, y):
    """Checks random field values x and y are independent."""
    self.assertEqual(x.ndim, 2)
    self.assertEqual(y.ndim, 2)
    # corr[i, j] = Correlation(x[:, i], y[:, j])
    corr = tfp.stats.correlation(x, y, sample_axis=0, event_axis=1)
    standard_error = 2 / np.sqrt(x.shape[0])  # product of two iid χ²
    np.testing.assert_array_less(corr, 4 * standard_error)


class GaussianRandomFieldTest(BaseRandomFieldTest):
  """Tests GaussianRandomField."""

  @parameterized.named_parameters(
      dict(
          testcase_name='T42_reasonable_corrs',
          variance=0.7,
          grid=spherical_harmonic.Grid.T42(),
          correlation_length=0.15,
          correlation_time=3,
      ),
      dict(
          testcase_name='T21_reasonable_corrs',
          variance=1.5,
          grid=spherical_harmonic.Grid.T21(),
          correlation_length=0.15,
          correlation_time=3,
      ),
      dict(
          testcase_name='T42_large_radius',
          variance=1.2,
          grid=spherical_harmonic.Grid.T42(radius=4),
          correlation_length=1.15,
          correlation_time=3,
      ),
      dict(
          testcase_name='T85_long_corrs',
          variance=2.7,
          grid=spherical_harmonic.Grid.T85(),
          correlation_length=0.5,
          correlation_time=3,
      ),
      dict(
          testcase_name='variance_None',
          variance=None,
          grid=spherical_harmonic.Grid.T21(),
          correlation_length=0.5,
          correlation_time=3,
      ),
  )
  def test_unconditional_and_trajectory_stats(
      self,
      variance,
      grid,
      correlation_length,
      correlation_time,
  ):
    coords = coordinate_systems.CoordinateSystem(
        grid, sigma_coordinates.SigmaCoordinates.equidistant(12)
    )
    aux_features = {}
    if variance is None:
      variance_arg = None
    else:
      variance_arg = variance * scales.units.dimensionless
    grf = stochastic.GaussianRandomField(
        coords,
        self.dt,
        self.physics_specs,
        aux_features,
        correlation_time=correlation_time * scales.units.hours,
        correlation_length=correlation_length * scales.units.dimensionless,
        variance=variance_arg,
    )
    self.assertEqual(
        grf.preferred_representation, stochastic.PreferredRepresentation.MODAL
    )
    self.check_unconditional_and_trajectory_stats(
        coords,
        grf,
        0.0,  # mean = 0
        variance,
        grid,
        correlation_length,
        correlation_time,
    )


class CenteredLognormalRandomFieldTest(BaseRandomFieldTest):
  """Tests CenteredLognormalRandomField."""

  @parameterized.named_parameters(
      dict(
          testcase_name='T42',
          grid=spherical_harmonic.Grid.T42(),
      ),
      dict(
          testcase_name='T21',
          grid=spherical_harmonic.Grid.T21(),
      ),
      dict(
          testcase_name='T42_large_radius',
          grid=spherical_harmonic.Grid.T42(radius=40),
      ),
      dict(
          testcase_name='T42_small_radius',
          grid=spherical_harmonic.Grid.T42(radius=0.5),
      ),
      dict(
          testcase_name='T85',
          grid=spherical_harmonic.Grid.T85(),
      ),
      dict(
          testcase_name='T85_small_radius',
          grid=spherical_harmonic.Grid.T85(radius=0.5),
      ),
  )
  def test_stats(self, grid):
    coords = coordinate_systems.CoordinateSystem(
        grid, sigma_coordinates.SigmaCoordinates.equidistant(12)
    )
    aux_features = {}

    variance = 3.0  # In the nonlinear regime.
    correlation_time = 1.1
    correlation_length = 0.15

    rf = stochastic.CenteredLognormalRandomField(
        coords,
        self.dt,
        self.physics_specs,
        aux_features,
        correlation_time=correlation_time,
        correlation_length=correlation_length,
        variance=variance,
    )
    self.assertEqual(
        rf.preferred_representation, stochastic.PreferredRepresentation.NODAL
    )
    self.check_unconditional_and_trajectory_stats(
        coords,
        rf,
        0.0,  # mean = 0 since our lognormals are centered.
        variance,
        grid,
        correlation_length,
        correlation_time,
        run_mean_check=True,
        run_variance_check=True,
        run_correlation_length_check=False,
        run_correlation_time_check=False,
        mean_tol_in_standard_errs=(
            # On coarser grids we will fail the pointwise tests due to the field
            # not being that homogeneous.
            # In all cases, the lognormal is heavy tailed, so we require looser
            # tolerance than the GaussianRandomField.
            10
            if grid.total_wavenumbers < 85
            else 5
        ),
        var_tol_in_standard_errs=(
            # The 4th moment of lognormal is quite large.
            # Nonetheless, due to the large number of effective samples
            # (considering 500 samples and averaging over the globe), the test
            # does indeed check something useful.
            80
            if grid.total_wavenumbers < 85
            else 9
        ),
    )


class BatchGaussianRandomFieldModuleTest(BaseRandomFieldTest):

  def setUp(self):
    super().setUp()
    grid = spherical_harmonic.Grid.T85()
    self.coords = coordinate_systems.CoordinateSystem(
        grid, sigma_coordinates.SigmaCoordinates.equidistant(12)
    )

  def _make_grf(
      self,
      variances,
      initial_correlation_lengths,
      initial_correlation_times,
      field_subset=None,
      n_fixed_fields=None,
  ):
    self.physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    self.dt = self.physics_specs.nondimensionalize(1 * scales.units.hour)

    return stochastic.BatchGaussianRandomFieldModule(
        self.coords,
        self.dt,
        self.physics_specs,
        aux_features={},
        initial_correlation_times=initial_correlation_times,
        initial_correlation_lengths=initial_correlation_lengths,
        variances=variances,
        field_subset=field_subset,
        n_fixed_fields=n_fixed_fields,
    )

  def nondimensionalize(self, x):
    return stochastic.nondimensionalize(x, self.physics_specs)

  @parameterized.named_parameters(
      dict(
          testcase_name='reasonable_corrs',
          variances=(1.0, 2.7),
          initial_correlation_lengths=(0.15, 0.2),
          initial_correlation_times=(1, 2.1),
      ),
      dict(
          testcase_name='one_fixed_field',
          variances=(1.0, 2.7),
          initial_correlation_lengths=(0.15, 0.2),
          initial_correlation_times=(1, 2.1),
          n_fixed_fields=1,
      ),
      dict(
          testcase_name='reasonable_corrs_skip_middle',
          # Using NaN for the one that should be skipped as a extra means to
          # test it wasn't used.
          variances=(1.0, np.nan, 2.7),
          initial_correlation_lengths=(0.5, np.nan, 0.2),
          initial_correlation_times=(1, np.nan, 2.1),
          field_subset=(0, 2),
      ),
  )
  def test_stats(
      self,
      variances,
      initial_correlation_lengths,
      initial_correlation_times,
      field_subset=None,
      n_fixed_fields=None,
  ):
    unroll_length = 10

    @hk.transform
    def make_field_trajectory(key):
      grf = self._make_grf(
          variances=variances,
          initial_correlation_lengths=initial_correlation_lengths,
          initial_correlation_times=initial_correlation_times,
          # Do not specify the field names... Let the default naming happen.
          field_subset=field_subset,
          n_fixed_fields=n_fixed_fields,
      )
      initial_value = grf.unconditional_sample(key)

      def step_fn(c, _):
        next_c = grf.advance(c)
        next_output = next_c.nodal_value
        return (next_c, next_output)

      _, trajectory = jax.lax.scan(
          step_fn, initial_value, xs=None, length=unroll_length
      )
      return initial_value, jax.device_get(trajectory)

    n_fixed_fields = n_fixed_fields or 0
    n_samples = 1000
    rngs = jax.random.split(jax.random.PRNGKey(802701), n_samples)
    params = make_field_trajectory.init(rng=rngs[0], key=rngs[0])
    initial_value, trajectory = jax.vmap(
        lambda rng: make_field_trajectory.apply(params, rng, rng)
    )(rngs)

    if field_subset:
      subset = lambda seq: [seq[i] for i in field_subset]
      variances = subset(variances)
      initial_correlation_times = subset(initial_correlation_times)
      initial_correlation_lengths = subset(initial_correlation_lengths)
    n_fields = len(variances)

    self.assertEqual(
        (n_samples, unroll_length, n_fields)
        + self.coords.horizontal.nodal_shape,
        trajectory.shape,
    )
    final_nodal_value = trajectory[:, -1]

    self.assertEqual(
        (n_samples, n_fields) + self.coords.horizontal.modal_shape,
        initial_value.core.shape,
    )
    self.assertEqual(
        (n_samples, n_fields) + self.coords.horizontal.modal_shape,
        initial_value.modal_value.shape,
    )
    self.assertEqual(
        (n_samples, n_fields) + self.coords.horizontal.nodal_shape,
        initial_value.nodal_value.shape,
    )

    # Check stochastic params were initialized with hk parameters.
    self.assertCountEqual(
        ['batch_gaussian_random_field_module'],
        params.keys(),
    )
    self.assertCountEqual(
        ['correlation_times_raw', 'correlation_lengths_raw'],
        params['batch_gaussian_random_field_module'].keys(),
    )
    for name in ['correlation_times_raw', 'correlation_lengths_raw']:
      self.assertEqual(
          (n_fields - n_fixed_fields,),
          params['batch_gaussian_random_field_module'][name].shape,
      )

    # Core should be modal
    tree_map(
        np.testing.assert_array_equal,
        initial_value.core,
        initial_value.modal_value,
    )

    # Nodal values should have the right statistics.
    for i, (variance, correlation_length) in enumerate(
        zip(
            variances,
            initial_correlation_lengths,
            strict=True,
        )
    ):
      for x in [initial_value.nodal_value, final_nodal_value]:
        self.check_mean(
            x[:, i],
            self.coords,
            expected_mean=0.0,
            variance=variance,
            correlation_length=correlation_length,
            mean_tol_in_standard_errs=5,
        )
        self.check_variance(
            x[:, i],
            self.coords,
            correlation_length=correlation_length,
            expected_variance=variance,
            var_tol_in_standard_errs=5,
        )
        self.check_correlation_length(
            x[:, i],
            expected_correlation_length=correlation_length,
            coords=self.coords,
        )

      # Fields 0 and 1 should be independent.
      # Check a handful of groups of nodal values.
      self.check_independent(
          x[:, 0, 100:105, 60],
          x[:, 1, 100:105, 60],
      )
      self.check_independent(
          x[:, 0, 100:105, 0],
          x[:, 1, 100:105, 0],
      )
      self.check_independent(x[:, 0, 0:5, 0], x[:, 1, 0:5, 0])

    # Initial and final sample should be independent as well, since we unroll
    # for much longer than the correlation time.
    self.check_independent(
        initial_value.nodal_value[:, 0, 50:55, 60],
        final_nodal_value[:, 0, 50:55, 60],
    )

  def test_giant_correlations_give_constant_fields(self):
    unroll_length = 10

    initial_correlation_lengths = [
        # Include a moderate correlation batch member just to check that extreme
        # correlations in other batch members don't mess this up.
        0.2,
        # Include the default "CONSTANT" correlations
        f'{CONSTANT_CORRELATION_LENGTH_KM} km',
        # Include a much larger correlation, to check for numerical stability.
        f'{1000 * CONSTANT_CORRELATION_LENGTH_KM} km',
    ]
    initial_correlation_times = [
        1,
        f'{CONSTANT_CORRELATION_TIME_HRS} hours',
        f'{1000 * CONSTANT_CORRELATION_TIME_HRS} hours',
    ]
    variances = [1.0, 1.0, 1.0]

    @hk.transform
    def make_field_trajectory(key):
      grf = self._make_grf(
          variances=variances,
          initial_correlation_lengths=initial_correlation_lengths,
          initial_correlation_times=initial_correlation_times,
          # Do not specify the field names... Let the default naming happen.
      )
      initial_value = grf.unconditional_sample(key)

      def step_fn(c, _):
        next_c = grf.advance(c)
        next_output = next_c.nodal_value
        return (next_c, next_output)

      _, trajectory = jax.lax.scan(
          step_fn, initial_value, xs=None, length=unroll_length
      )
      return initial_value, jax.device_get(trajectory)

    n_samples = 100
    rngs = jax.random.split(jax.random.PRNGKey(802701), n_samples)
    params = make_field_trajectory.init(rng=rngs[0], key=rngs[0])
    initial_value, trajectory = jax.vmap(
        lambda rng: make_field_trajectory.apply(
            params, rng, jax.random.fold_in(rng, 1)
        )
    )(rngs)
    final_nodal_value = trajectory[:, -1]
    initial_nodal_value = initial_value.nodal_value

    self.assertTrue(np.all(np.isfinite(initial_nodal_value)))
    self.assertTrue(np.all(np.isfinite(final_nodal_value)))

    # All fields have the correct mean and variance.
    for i in range(3):
      for x in [initial_value.nodal_value, final_nodal_value]:
        self.check_mean(
            x[:, i],
            self.coords,
            expected_mean=0.0,
            variance=variances[i],
            correlation_length=self.nondimensionalize(
                initial_correlation_lengths[i]
            ),
            mean_tol_in_standard_errs=5,
        )
        self.check_variance(
            x[:, i],
            self.coords,
            correlation_length=self.nondimensionalize(
                initial_correlation_lengths[i]
            ),
            expected_variance=variances[i],
            var_tol_in_standard_errs=5,
        )

    # Field 0 (moderate correlation) has correct correlation length.
    for x in [initial_value.nodal_value, final_nodal_value]:
      self.check_correlation_length(
          x[:, 0],
          expected_correlation_length=initial_correlation_lengths[0],
          coords=self.coords,
      )

    # The variation in index 0 (moderate correlation) is much larger than that
    # in index 1 (CONSTANT_CORRELATION_*) or 2.
    # This checks the correlation length/time of fields 1, 2 is huge.
    diff = final_nodal_value - initial_nodal_value
    for i in [1, 2]:
      self.assertGreater(  # Variation in time
          np.max(np.abs(diff[:, 0])), 100 * np.max(np.abs(diff[:, i]))
      )
      self.assertGreater(  # Variation in lat/lon
          np.std(initial_nodal_value[:, 0], axis=(-1, -2)).max(),
          10000 * np.std(initial_nodal_value[:, i], axis=(-1, -2)).max(),
      )


class DictOfGaussianRandomFieldModulesTest(BaseRandomFieldTest):

  def setUp(self):
    super().setUp()
    grid = spherical_harmonic.Grid.T85()
    self.coords = coordinate_systems.CoordinateSystem(
        grid, sigma_coordinates.SigmaCoordinates.equidistant(12)
    )

  def _make_grf(
      self,
      variances,
      initial_correlation_lengths,
      initial_correlation_times,
      field_names=None,
      field_subset=None,
  ):
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    self.dt = physics_specs.nondimensionalize(1 * scales.units.hour)

    return stochastic.DictOfGaussianRandomFieldModules(
        self.coords,
        self.dt,
        physics_specs,
        aux_features={},
        initial_correlation_times=initial_correlation_times,
        initial_correlation_lengths=initial_correlation_lengths,
        variances=variances,
        field_names=field_names,
        field_subset=field_subset,
    )

  @parameterized.named_parameters(
      dict(
          testcase_name='reasonable_corrs',
          variances=(1.0, 2.7),
          initial_correlation_lengths=(0.15, 0.25),
          initial_correlation_times=(2, 3),
      ),
      dict(
          testcase_name='reasonable_corrs_skip_middle',
          # Using NaN for the one that should be skipped as a extra means to
          # test it wasn't used.
          variances=(1.0, np.nan, 2.7),
          initial_correlation_lengths=(0.15, np.nan, 0.25),
          initial_correlation_times=(2, np.nan, 3),
          field_subset=(0, 2),
      ),
  )
  def test_initialization_and_stats(
      self,
      variances,
      initial_correlation_lengths,
      initial_correlation_times,
      field_subset=None,
  ):
    @hk.transform
    def unconditional_sample(key):
      grf = self._make_grf(
          variances=variances,
          initial_correlation_lengths=initial_correlation_lengths,
          initial_correlation_times=initial_correlation_times,
          # Do not specify the field names... Let the default naming happen.
          field_subset=field_subset,
      )
      return grf.unconditional_sample(key)

    n_samples = 500
    rngs = jax.random.split(jax.random.PRNGKey(802701), n_samples)
    params = unconditional_sample.init(rng=rngs[0], key=rngs[0])
    sample = jax.vmap(lambda rng: unconditional_sample.apply(params, rng, rng))(
        rngs
    )

    # Now that we are done using the GRF, trim the args to the length of args
    # that *should* have been used (in light of subsetting).
    field_names = [f'GRF{i}' for i in range(len(variances))]
    if field_subset:
      subset = lambda seq: [seq[i] for i in field_subset]
      field_names = subset(field_names)
      variances = subset(variances)
      initial_correlation_times = subset(initial_correlation_times)
      initial_correlation_lengths = subset(initial_correlation_lengths)

    # Check stochastic params were initialized with hk parameters.
    self.assertCountEqual(
        params.keys(),
        [
            f'dict_of_gaussian_random_field_modules/~/{name}'
            for name in field_names
        ],
    )
    for one_rf_name, one_rf_params in params.items():
      self.assertCountEqual(
          # Correlations are tuned, variance is fixed and so doesn't show up
          # here.
          ['correlation_time_raw', 'correlation_length_raw'],
          one_rf_params.keys(),
          msg=f'Failed at {one_rf_name=}',
      )

    # Core should be modal
    tree_map(np.testing.assert_array_equal, sample.core, sample.modal_value)

    # Nodal values should have the right statistics.
    for name, variance, correlation_length in zip(
        field_names,
        variances,
        initial_correlation_lengths,
        strict=True,
    ):
      self.check_mean(
          sample.nodal_value[name],
          self.coords,
          expected_mean=0.0,
          variance=variance,
          correlation_length=correlation_length,
          mean_tol_in_standard_errs=5,
      )
      self.check_variance(
          sample.nodal_value[name],
          self.coords,
          correlation_length=correlation_length,
          expected_variance=variance,
          var_tol_in_standard_errs=5,
      )
      self.check_correlation_length(
          sample.nodal_value[name],
          expected_correlation_length=correlation_length,
          coords=self.coords,
      )

    # Different fields should be independent.
    self.check_independent(
        sample.nodal_value[field_names[0]][:, 100:105, 64],
        sample.nodal_value[field_names[1]][:, 100:105, 64],
    )


class SumOfGaussianRandomFieldsTest(BaseRandomFieldTest):

  def test_sum_of_two_gaussians_that_only_differ_in_variance(self):
    # The sum of two independent RFs with the same correlation lengths results
    # in a RF with the sum of variance, but the same correlation lengths.
    grid = spherical_harmonic.Grid.T21()
    coords = coordinate_systems.CoordinateSystem(
        grid, sigma_coordinates.SigmaCoordinates.equidistant(12)
    )
    aux_features = {}

    correlation_time = 2.1
    correlation_length = 0.15
    variances = [1.0, 2.0]

    sum_rf = stochastic.SumOfGaussianRandomFields(
        coords,
        self.dt,
        self.physics_specs,
        aux_features,
        correlation_times=[correlation_time] * 2,
        correlation_lengths=[correlation_length] * 2,
        variances=variances,
    )

    self.assertEqual(
        sum_rf.preferred_representation,
        stochastic.PreferredRepresentation.MODAL,
    )

    self.check_unconditional_and_trajectory_stats(
        coords,
        sum_rf,
        0.0,  # mean = 0
        sum(variances),
        grid,
        correlation_length,
        correlation_time,
    )


class GaussianRandomFieldModuleTest(parameterized.TestCase):

  def _make_module_and_params_and_grf(
      self,
      initial_variance='0.5 dimensionless',
      variance_bound=None,
      tune_variance=True,
  ):
    """Makes hk.Module, and params / Loss."""
    coords = coordinate_systems.CoordinateSystem(
        spherical_harmonic.Grid.T21(),
        sigma_coordinates.SigmaCoordinates.equidistant(12),
    )
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    self.dt = physics_specs.nondimensionalize(1 * scales.units.hour)

    @hk.transform
    def f():
      return stochastic.GaussianRandomFieldModule(
          coords,
          self.dt,
          physics_specs,
          aux_features={},
          initial_correlation_time='3 dimensionless',
          initial_correlation_length='50 km',
          initial_variance=initial_variance,
          variance_bound=variance_bound,
          tune_variance=tune_variance,
      )

    params = f.init(rng=None)
    grf = f.apply(params, rng=None)
    return f, params, grf

  @parameterized.named_parameters(
      dict(
          testcase_name='Variance0p5',
          initial_variance='0.5 dimensionless',
          expected_variance=0.5,
      ),
      dict(
          testcase_name='VarianceNone',
          initial_variance=None,
          expected_variance=None,
      ),
      dict(
          testcase_name='Variance0p5Bounded',
          initial_variance='0.5 dimensionless',
          expected_variance=0.5,
          variance_bound='1 dimensionless',
      ),
      dict(
          testcase_name='Variance0p5DoNotTune',
          initial_variance=0.5,
          expected_variance=0.5,
          tune_variance=False,
      ),
  )
  def test_parameter_initialization(
      self,
      initial_variance,
      expected_variance,
      variance_bound=None,
      tune_variance=True,
  ):
    _, params, grf = self._make_module_and_params_and_grf(
        initial_variance=initial_variance,
        variance_bound=variance_bound,
        tune_variance=tune_variance,
    )

    expected_raw_params = ['correlation_time_raw', 'correlation_length_raw']
    if tune_variance:
      expected_raw_params += ['variance_raw']

    self.assertCountEqual(['gaussian_random_field_module'], params)
    self.assertCountEqual(
        expected_raw_params,
        params['gaussian_random_field_module'],
    )
    self.assertEqual(
        grf.preferred_representation, stochastic.PreferredRepresentation.MODAL
    )

    np.testing.assert_array_equal(expected_variance, grf.variance)

    # C.f. correlation_time = '3 dimensionless'
    np.testing.assert_allclose(np.exp(-self.dt / 3), grf.phi)

    n_samples = 500
    init_rngs = jax.random.split(jax.random.PRNGKey(5), n_samples)
    initial_states = jax.vmap(grf.unconditional_sample)(init_rngs)
    initial_values = jax.vmap(grf.to_nodal_values)(initial_states.core)

    # Test stats...these are tested more carefully by
    # CenteredLognormalRandomFieldTest, but let's double check here.
    np.testing.assert_allclose(0.0, initial_values.mean(), atol=1e-2)
    np.testing.assert_allclose(
        expected_variance or 0.0,
        initial_values.var(axis=0).mean(),
        rtol=1e-2,
        atol=1e-5,
    )


class CenteredLognormalRandomFieldModuleTest(parameterized.TestCase):

  def _make_module_and_params_and_rf(
      self,
      initial_variance=0.5,
      variance_bound=None,
  ):
    """Makes hk.Module, and params / Loss."""
    coords = coordinate_systems.CoordinateSystem(
        spherical_harmonic.Grid.T21(),
        sigma_coordinates.SigmaCoordinates.equidistant(12),
    )
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    self.dt = physics_specs.nondimensionalize(1 * scales.units.hour)

    @hk.transform
    def f():
      return stochastic.CenteredLognormalRandomFieldModule(
          coords,
          self.dt,
          physics_specs,
          aux_features={},
          initial_correlation_time='1 dimensionless',
          initial_correlation_length='5km',
          initial_variance=initial_variance,
          variance_bound=variance_bound,
      )

    params = f.init(rng=None)
    rf = f.apply(params, rng=None)
    return f, params, rf

  @parameterized.named_parameters(
      dict(
          testcase_name='Variance0p5',
          initial_variance=0.5,
          expected_variance=0.5,
      ),
      dict(
          testcase_name='VarianceNone',
          initial_variance=None,
          expected_variance=None,
      ),
      dict(
          testcase_name='Variance0p5Bounded',
          initial_variance=0.5,
          expected_variance=0.5,
          variance_bound=1,
      ),
  )
  def test_parameter_initialization(
      self,
      initial_variance,
      expected_variance,
      variance_bound=None,
  ):
    _, params, rf = self._make_module_and_params_and_rf(
        initial_variance=initial_variance,
        variance_bound=variance_bound,
    )

    self.assertCountEqual(['centered_lognormal_random_field_module'], params)
    self.assertCountEqual(
        ['correlation_time_raw', 'correlation_length_raw', 'variance_raw'],
        params['centered_lognormal_random_field_module'],
    )
    self.assertEqual(
        rf.preferred_representation, stochastic.PreferredRepresentation.NODAL
    )

    np.testing.assert_array_equal(expected_variance, rf.variance)

    # C.f. correlation_time = '1 dimensionless'
    np.testing.assert_allclose(np.exp(-self.dt / 1), rf.phi)

    n_samples = 500
    init_rngs = jax.random.split(jax.random.PRNGKey(5), n_samples)
    initial_states = jax.vmap(rf.unconditional_sample)(init_rngs)
    initial_values = jax.vmap(rf.to_nodal_values)(initial_states.core)

    # Recall this is Exp(ξ) - 1
    self.assertGreater(initial_values.min(), -1)

    # Test stats...these are tested more carefully by
    # CenteredLognormalRandomFieldTest, but let's double check here.
    np.testing.assert_allclose(0.0, initial_values.mean(), atol=1e-3)
    np.testing.assert_allclose(
        expected_variance or 0.0,
        initial_values.var(axis=0).mean(),
        rtol=1e-2,
        atol=1e-5,
    )


class SumOfGaussianRandomFieldsModuleTest(parameterized.TestCase):

  def _make_module_and_params_and_grf(
      self,
      initial_correlation_times,
      initial_correlation_lengths,
      initial_variances,
      variance_bounds,
  ):
    """Makes hk.Module, and params / Loss."""
    self.coords = coordinate_systems.CoordinateSystem(
        spherical_harmonic.Grid.T21(),
        sigma_coordinates.SigmaCoordinates.equidistant(12),
    )
    physics_specs = primitive_equations.PrimitiveEquationsSpecs.from_si()
    self.dt = physics_specs.nondimensionalize(1 * scales.units.hour)

    @hk.transform
    def f():
      return stochastic.SumOfGaussianRandomFieldsModule(
          self.coords,
          self.dt,
          physics_specs,
          aux_features={},
          initial_correlation_times=initial_correlation_times,
          initial_correlation_lengths=initial_correlation_lengths,
          initial_variances=initial_variances,
          variance_bounds=variance_bounds,
      )

    params = f.init(rng=None)
    grf = f.apply(params, rng=None)
    return f, params, grf

  @parameterized.named_parameters(
      dict(testcase_name='NoBound', variance_bound=None),
      dict(testcase_name='YesBound', variance_bound=5),
  )
  def test_parameter_initialization(self, variance_bound):
    initial_correlation_times = [1.0, 3.0]
    initial_correlation_lengths = [2.0, 4]
    initial_variances = [0.5, 1.5]
    _, params, grf = self._make_module_and_params_and_grf(
        initial_correlation_times,
        initial_correlation_lengths,
        initial_variances,
        variance_bounds=[variance_bound] * 2,
    )

    self.assertCountEqual(
        [
            'sum_of_gaussian_random_fields_module/~/gaussian_random_field_module',
            'sum_of_gaussian_random_fields_module/~/gaussian_random_field_module_1',
        ],
        params,
    )
    for top_key in params:
      self.assertCountEqual(
          ['correlation_time_raw', 'correlation_length_raw', 'variance_raw'],
          params[top_key],
      )
    self.assertEqual(
        grf.preferred_representation, stochastic.PreferredRepresentation.MODAL
    )

    for i, grf_i in enumerate(grf._random_fields):
      # Variance may not match exactly, if the bound is used.
      np.testing.assert_allclose(initial_variances[i], grf_i.variance)
      self.assertEqual(
          np.exp(-self.dt / initial_correlation_times[i]), grf_i.phi
      )
      self.assertEqual(
          (initial_correlation_lengths[i] / self.coords.horizontal.radius) ** 2
          / 2,
          grf_i.kt,
      )


class ParameterConversionTest(parameterized.TestCase):

  def test_convert_hk_param_to_positive_scalar(self):
    initial_value = 0.9

    def convert(param):
      return stochastic.convert_hk_param_to_positive_scalar(
          param, initial_value=initial_value
      )

    np.testing.assert_allclose(convert(0.0), initial_value)
    np.testing.assert_allclose(convert(-20.0), 0.0, atol=1e-7)
    for x in np.linspace(-2.0, 2.0, 10):
      self.assertGreater(convert(x), 0.0)

  def test_convert_hk_param_to_bounded_scalar(self):
    low = 0.1
    high = 2.2
    initial_value = 0.9

    def convert(param):
      return stochastic.convert_hk_param_to_bounded_scalar(
          param, initial_value=initial_value, low=low, high=high
      )

    np.testing.assert_allclose(convert(0.0), initial_value)
    np.testing.assert_allclose(convert(-20.0), low)
    np.testing.assert_allclose(convert(20.0), high)
    for x in np.linspace(-2.0, 2.0, 10):
      self.assertBetween(convert(x), low, high)


if __name__ == '__main__':
  jax.config.parse_flags_with_absl()
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from absl.testing import absltest
from absl.testing import parameterized
import jax
from neuralgcm import optimization
import numpy as np
import optax


class PiecewiseConstantScheduleTest(parameterized.TestCase):

  def test_3_leg_schedule(self):
    params = {'theta': 0.0}
    gradients = jax.tree_util.tree_map(np.ones_like, params)

    schedule = optimization.piecewise_constant_schedule_specified_by_rates(
        rates=[0.0, 1.0, 10.0],
        boundaries=[2, 4],
    )
    optimizer = optax.sgd(schedule)
    opt_state = optimizer.init(params)

    updates, opt_state = optimizer.update(gradients, opt_state, params)
    params = optax.apply_updates(params, updates)
    self.assertEqual({'theta': 0.0}, params)  # 0 - 0*1 = 0

    updates, opt_state = optimizer.update(gradients, opt_state, params)
    params = optax.apply_updates(params, updates)
    self.assertEqual({'theta': 0.0}, params)  # 0 - 0*1 = 0

    updates, opt_state = optimizer.update(gradients, opt_state, params)
    params = optax.apply_updates(params, updates)
    self.assertEqual({'theta': -1.}, params)  # 0 - 1*1 = -1

    updates, opt_state = optimizer.update(gradients, opt_state, params)
    params = optax.apply_updates(params, updates)
    self.assertEqual({'theta': -2.}, params)  # -1 - 1*1 = -2

    updates, opt_state = optimizer.update(gradients, opt_state, params)
    params = optax.apply_updates(params, updates)
    self.assertEqual({'theta': -12.}, params)  # -2 - 10*1 = -12


class TopLevelMultiRateTest(parameterized.TestCase):

  def _make_exponential_decay(self, lr: float):
    return optax.warmup_exponential_decay_schedule(
        init_value=lr,
        peak_value=2 * lr,
        warmup_steps=2,
        transition_steps=2,
        decay_rate=0.1,
    )

  @parameterized.named_parameters(
      dict(testcase_name='Default'),
      dict(testcase_name='ExponentialDecay', use_exponential_decay=True),
  )
  def test_rates_adjusted_well(self, use_exponential_decay=False):
    if use_exponential_decay:
      make_lr = self._make_exponential_decay
    else:
      make_lr = lambda lr: lr

    params = {
        'Top_A': {'w': 0.0, 'b': 0.0},
        'Top_B': {'w': 0.0, 'b': 0.0},
        'my_cats_breath': {'w': 0.0, 'b': 0.0},
        'cats_kill_birds': {'w': 5.0, 'b': 8.0},
    }
    gradients = jax.tree_util.tree_map(np.ones_like, params)

    # Give Top_A lr=100, cats lr=1, and all others lr=1e-3
    optimizer = optimization.top_level_multi_adam(
        top_level_keys=['Top_A', 'REGEX_cats'],
        learning_rates=[make_lr(100), make_lr(1)],
        default_learning_rate=make_lr(1e-3),
    )
    state = optimizer.init(params)
    updates, unused_new_state = optimizer.update(gradients, state, params)
    new_params = optax.apply_updates(params, updates)

    # We took one step, so the final parameter values should be the initial
    # minus 1x learning rate (up to some Adam initialization noise).
    np.testing.assert_allclose(0 - 100, new_params['Top_A']['w'], rtol=1e-5)
    np.testing.assert_allclose(0 - 100, new_params['Top_A']['b'], rtol=1e-5)
    np.testing.assert_allclose(
        0 - 1, new_params['my_cats_breath']['w'], rtol=1e-5
    )
    np.testing.assert_allclose(
        0 - 1, new_params['my_cats_breath']['b'], rtol=1e-5
    )
    np.testing.assert_allclose(
        5 - 1, new_params['cats_kill_birds']['w'], rtol=1e-5
    )
    np.testing.assert_allclose(
        8 - 1, new_params['cats_kill_birds']['b'], rtol=1e-5
    )
    np.testing.assert_allclose(0 - 1e-3, new_params['Top_B']['w'], rtol=1e-5)
    np.testing.assert_allclose(0 - 1e-3, new_params['Top_B']['b'], rtol=1e-5)

  def test_missing_keys_handling(self):
    params = {'Top_A': {'w': 0.0, 'b': 0.0}, 'Top_B': {'w': 0.0, 'b': 0.0}}

    with self.subTest('Raises if raise_if_keys_not_found=True'):
      optimizer = optimization.top_level_multi_adam(
          top_level_keys=['Top_A', 'NOT_IN_PARAMS'],
          learning_rates=[1, 2],
          raise_if_keys_not_found=True,
      )
      with self.assertRaisesRegex(optimization.OptimizerError, 'NOT_IN_PARAMS'):
        optimizer.init(params)

    with self.subTest('Does not raise if raise_if_keys_not_found=False'):
      optimizer = optimization.top_level_multi_adam(
          top_level_keys=['Top_A', 'NOT_IN_PARAMS'],
          learning_rates=[1, 2],
          raise_if_keys_not_found=False,
      )
      optimizer.init(params)  # Should not raise


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Modules that computes relevant state features to be used by ML components."""

from typing import Any, Callable, Mapping, Optional, Protocol, Sequence
from dinosaur import coordinate_systems
from dinosaur import primitive_equations
from dinosaur import pytree_utils
from dinosaur import radiation
from dinosaur import scales
from dinosaur import spherical_harmonic
from dinosaur import typing
from dinosaur import xarray_utils
import gin
import haiku as hk
import jax
import jax.numpy as jnp
from neuralgcm import transforms
import numpy as np


Array = typing.Array
Pytree = typing.Pytree
TransformModule = typing.TransformModule
KeyWithCosLatFactor = typing.KeyWithCosLatFactor


class FeaturesFn(Protocol):

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> typing.Pytree:
    ...


FeaturesModule = Callable[..., FeaturesFn]


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class PrimitiveEquationsDiagnosticState(hk.Module):
  """Features modules that returns processed DiagnosticState for PE."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      features_transform_module: TransformModule = transforms.IdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.features_transform_fn = features_transform_module(
        coords, dt, physics_specs, aux_features
    )
    self.coords = coords

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> primitive_equations.DiagnosticState:
    del memory, diagnostics, randomness, forcing  # unused
    if not isinstance(inputs, primitive_equations.State):
      inputs = primitive_equations.State(**inputs)
    d_state = primitive_equations.compute_diagnostic_state(inputs, self.coords)
    return self.features_transform_fn(d_state.asdict())


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class VelocityAndPrognostics(hk.Module):
  """Features module that returns prognostics + u,v and optionally gradients."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      fields_to_include: Optional[Sequence[str]] = None,
      features_transform_module: TransformModule = transforms.IdentityTransform,
      compute_gradients_module: TransformModule = transforms.EmptyTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.features_transform_fn = features_transform_module(
        coords, dt, physics_specs, aux_features
    )
    self.coords = coords
    self.fields_to_include = fields_to_include
    self.compute_gradients_fn = compute_gradients_module(
        coords, dt, physics_specs, aux_features
    )

  def _extract_features(
      self,
      inputs: typing.Pytree,
      prefix: str = '',
  ) -> typing.Pytree:
    """Returns a nodal velocity and prognostic features."""
    # Note: all intermediate features have an explicit cos-lat factors in key.
    # These factors are removed in the `__call__` method before returning.

    # compute `u, v` if div/curl is available and `u, v` not in prognosics.
    if set(['vorticity', 'divergence']).issubset(inputs.keys()) and not set(
        ['u', 'v']
    ).intersection(inputs.keys()):
      cos_lat_u, cos_lat_v = spherical_harmonic.get_cos_lat_vector(
          inputs['vorticity'], inputs['divergence'], self.coords.horizontal
      )
      modal_features = {
          KeyWithCosLatFactor(prefix + 'u', 1): cos_lat_u,
          KeyWithCosLatFactor(prefix + 'v', 1): cos_lat_v,
      }
    else:
      modal_features = {}
    prognostics_keys = list(inputs.keys())
    prognostics_keys.remove('tracers')
    prognostics_keys.remove('sim_time')
    for k in prognostics_keys:
      if self.fields_to_include is None or k in self.fields_to_include:
        modal_features[KeyWithCosLatFactor(prefix + k, 0)] = inputs[k]

    for k, v in inputs['tracers'].items():
      if self.fields_to_include is None or k in self.fields_to_include:
        modal_features[KeyWithCosLatFactor(prefix + k, 0)] = v
    # Computing gradient features and adjusting cos_lat factors.
    modal_features = self.coords.with_dycore_sharding(modal_features)
    diff_operator_features = self.compute_gradients_fn(modal_features)
    sec_lat = 1 / self.coords.horizontal.cos_lat
    sec2_lat = self.coords.horizontal.sec2_lat
    sec_lat_scales = {0: 1, 1: sec_lat, 2: sec2_lat}
    # Computing all features in nodal space.
    features = {}
    for k, v in (diff_operator_features | modal_features).items():
      sec_lat_scale = sec_lat_scales[k.factor_order]
      features[k.name] = self.coords.horizontal.to_nodal(v) * sec_lat_scale
    features = self.coords.with_dycore_sharding(features)
    return features

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> typing.Pytree:
    del memory, diagnostics, randomness, forcing  # unused.
    nodal_features = self._extract_features(inputs)
    return self.features_transform_fn(nodal_features)


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class MemoryVelocityAndValues(VelocityAndPrognostics):
  """Similar to `VelocityAndPrognostics`, but operates on memory."""

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> typing.Pytree:
    del inputs, diagnostics, randomness, forcing  # unused.
    nodal_features = self._extract_features(memory, 'memory_')
    return self.features_transform_fn(nodal_features)


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class NodalInputVelocityAndPrognostics(VelocityAndPrognostics):
  """Features modules that returns velocities, temperature, and optionally gradients."""

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> typing.Pytree:
    to_modal_fn = self.coords.horizontal.to_modal
    inputs = to_modal_fn(inputs)
    memory = to_modal_fn(memory)
    return super().__call__(inputs, memory, randomness, forcing)


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class RadiationFeatures(hk.Module):
  """Feature module that computes incident radiation flux."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      features_transform_module: TransformModule = transforms.IdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.features_transform_fn = features_transform_module(
        coords, dt, physics_specs, aux_features
    )
    ref_datetime_str = str(aux_features[xarray_utils.REFERENCE_DATETIME_KEY])
    self.solar_radiation = radiation.SolarRadiation.normalized(
        coords=coords,
        physics_specs=physics_specs,
        reference_datetime=np.datetime64(ref_datetime_str),
    )

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> typing.Pytree:
    del memory, diagnostics, randomness, forcing  # unused.
    features = {}
    features['radiation'] = self.solar_radiation.radiation_flux(
        inputs['sim_time']
    )
    # TODO(janniyuval) add a flag that allow to get radiation of next time step
    # insert a feature axis.
    features = jax.tree_util.tree_map(lambda x: jnp.expand_dims(x, 0), features)
    return self.features_transform_fn(features)


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class OrbitalTimeFeatures(hk.Module):
  """Feature module that computes orbital time features."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      features_transform_module: TransformModule = transforms.IdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.features_transform_fn = features_transform_module(
        coords, dt, physics_specs, aux_features
    )
    ref_datetime_str = str(aux_features[xarray_utils.REFERENCE_DATETIME_KEY])
    self.solar_radiation = radiation.SolarRadiation.normalized(
        coords=coords,
        physics_specs=physics_specs,
        reference_datetime=np.datetime64(ref_datetime_str),
    )

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> typing.Pytree:
    del memory, diagnostics, randomness, forcing  # unused.
    features = {}
    # Cosine and sine of Earth's orbital phase around the Sun
    orbital_time = self.solar_radiation.time_to_orbital_time(inputs['sim_time'])
    # Convert from orbital_phase=0 on January 1st UTC to orbital_phase=0 at the
    # approximate perihelion (when earth is closest to the sun).
    orbital_phase = orbital_time.orbital_phase - radiation.PERIHELION
    # All longitude, latitude locations share the same orbital phase
    ones = jnp.ones(self.solar_radiation.coords.surface_nodal_shape)
    features['cos_orbital_phase'] = jnp.cos(orbital_phase) * ones
    features['sin_orbital_phase'] = jnp.sin(orbital_phase) * ones
    # Cosine and sine of local hour angle (angle from solar noon)
    solar_hour_angle = self.solar_radiation.solar_hour_angle(inputs['sim_time'])
    solar_hour_angle = jnp.expand_dims(solar_hour_angle, 0)
    features['cos_solar_hour'] = jnp.cos(solar_hour_angle)
    features['sin_solar_hour'] = jnp.sin(solar_hour_angle)
    # TODO(janniyuval) add a flag that allow to get radiation of next time step
    return self.features_transform_fn(features)


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class ForcingFeatures(hk.Module):
  """Feature module that provides forcing values as features."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      forcing_to_include: Sequence[str] = tuple(),
      features_transform_module: TransformModule = transforms.IdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.forcing_to_include = forcing_to_include
    self.features_transform_fn = features_transform_module(
        coords, dt, physics_specs, aux_features
    )

  def __call__(
      self,
      inputs: Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Forcing] = None,
  ) -> Pytree:
    del inputs, memory, diagnostics, randomness
    features = {}
    for key in self.forcing_to_include:
      value = forcing[key]
      # Expect singleton "level" dimension for surface forcings
      if value.ndim > 3:
        raise ValueError(
            f'Expected forcing "{key}" to have ndim <= 3, got {value.ndim}'
        )
      if value.ndim == 2:
        value = jnp.expand_dims(value, axis=0)
      if value.shape[0] != 1:
        raise ValueError(
            f'Expected forcing "{key}" to have leading dimension 1'
            f'for level, got {value.shape}'
        )
      features[key] = value

    return self.features_transform_fn(features)


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class LatitudeFeatures(hk.Module):
  """Feature module that creates cos and sin of latitude as features."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      features_transform_module: TransformModule = transforms.IdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.features_transform_fn = features_transform_module(
        coords, dt, physics_specs, aux_features
    )
    self.coords = coords

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> typing.Pytree:
    del inputs, memory, diagnostics, randomness, forcing  # unused.
    _, sin_lat = self.coords.horizontal.nodal_mesh
    sin_features = sin_lat[np.newaxis, ...]
    cos_features = jnp.cos(jnp.arcsin(sin_features))
    features = {
        'cos_latitude': cos_features,
        'sin_latitude': sin_features,
    }
    return self.features_transform_fn(features)


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class RandomnessFeatures(hk.Module):
  """Feature module that returns fields from `randomness` as features."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      features_transform_module: TransformModule = transforms.IdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.features_transform_fn = features_transform_module(
        coords, dt, physics_specs, aux_features
    )

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> typing.Pytree:
    del inputs, memory, diagnostics, forcing  # unused.
    if randomness is None:
      random_features = {}
    elif isinstance(randomness, dict):
      random_features, _ = pytree_utils.flatten_dict(randomness)
    elif isinstance(randomness, jax.Array):
      random_features = {'randomness': randomness}
    else:
      raise ValueError(f'randomness has unsupported {type(randomness)=}.')
    # random fields are 2D by construction, adding a feature/level dimension.
    if randomness is not None:
      ndims = set(x.ndim for x in jax.tree_util.tree_leaves(random_features))
      if not ndims.issubset({2, 3}):
        raise ValueError(
            f'Random fields expected to be 2D and/or 3D. Found {ndims=}'
        )

      def make_3d(x):
        if x.ndim == 3:
          return x
        if x.ndim == 2:
          return x[np.newaxis, ...]

      random_features = jax.tree_util.tree_map(make_3d, random_features)
    return self.features_transform_fn(random_features)


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class OrographyFeatures(hk.Module):
  """Feature module that computes orographic features."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      features_transform_module: TransformModule = transforms.IdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    if xarray_utils.OROGRAPHY not in aux_features:
      raise ValueError('OrographyFeatures requires orography in aux_features.')
    self.nodal_orography = aux_features[xarray_utils.OROGRAPHY]
    self.features_transform_fn = features_transform_module(
        coords, dt, physics_specs, aux_features
    )

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> typing.Pytree:
    del inputs, memory, diagnostics, randomness, forcing  # unused.
    features = {
        xarray_utils.OROGRAPHY: jnp.expand_dims(self.nodal_orography, 0),
    }
    return self.features_transform_fn(features)


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class OneHotAuxFeatures(hk.Module):
  """Feature module that produces one-hot encodings from binary covariates."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      covariate_keys: Sequence[str] = gin.REQUIRED,
      convert_float_to_int: bool = False,
      name: Optional[str] = None,
  ):
    del coords, dt, physics_specs  # unused.
    super().__init__(name=name)
    covariates = {}
    num_classes = {}
    for key in covariate_keys:
      if key not in aux_features:
        raise ValueError(f'Covariate {key} not found in aux_features.')
      if not np.issubdtype(aux_features[key].dtype, np.integer):
        if convert_float_to_int:
          aux_features[key] = np.round(aux_features[key]).astype(int)
        else:
          raise ValueError(
              f'Covariate {key} is expected to be integer dtype, '
              f'but is: {aux_features[key].dtype}'
          )
      covariates[key] = aux_features[key]
      num_classes[key] = np.unique(aux_features[key]).size
    self.covariates = covariates
    self.num_classes = num_classes

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> dict[str, jnp.ndarray]:
    del inputs, memory, diagnostics, randomness, forcing  # unused.
    features = {
        k: jax.nn.one_hot(v, self.num_classes[k], axis=0)
        for k, v in self.covariates.items()
    }
    return features


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class LearnedPositionalFeatures(hk.Module):
  """Feature module with learned params at surface nodal locations."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      latent_size: int,
      scale: float = 1.0,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.scale = scale
    self.padding = coords.horizontal.nodal_padding
    unpadded_nodal_shape = tuple(
        x - y for x, y in zip(coords.horizontal.nodal_shape, self.padding)
    )
    self.positional_features = hk.get_parameter(
        'learned_positional_features',
        (latent_size,) + unpadded_nodal_shape,
        jnp.float32,
        init=hk.initializers.Constant(0.0),
    )

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> dict[str, jnp.ndarray]:
    """Returns scaled parameter values at surface nodal locations."""
    del inputs, memory, diagnostics, randomness, forcing  # unused.
    pad_x, pad_y = self.padding
    positional_features = self.scale * jnp.pad(
        self.positional_features, [(0, 0), (0, pad_x), (0, pad_y)]
    )
    return {'learned_positional_features': positional_features}


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class EmbeddingSurfaceFeatures(hk.Module):
  """Feature module that specifies embedding surface outputs as features.

  Returns {feature_name: nn_output}
    where nn_output.shape = (output_size, lon, lat).
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      feature_name: str,
      output_size: int,
      embedding_module: typing.EmbeddingModule,
      features_transform_module: TransformModule = transforms.IdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    # output shapes are arrays to be pytree leaves for tree_map
    output_shapes = {
        feature_name: np.asarray((output_size,) + coords.horizontal.nodal_shape)
    }
    self.embedding_fn = embedding_module(
        coords, dt, physics_specs, aux_features, output_shapes=output_shapes
    )
    self.features_transform_fn = features_transform_module(
        coords, dt, physics_specs, aux_features
    )

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> typing.Pytree:
    features = self.embedding_fn(
        inputs, memory, diagnostics, randomness, forcing
    )
    return self.features_transform_fn(features)


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class EmbeddingVolumeFeatures(hk.Module):
  """Feature module that specifies embedding volume outputs as features.

  Returns {feature_name_0: nn_output_0,
           feature_name_1: nn_output_1,
           ...
          }
    where the NN output array has shape (output_size, level, lon, lat), which is
    unpacked over output_size such that nn_output_{i}.shape = (level, lon, lat)
    for each i in range(output_size).
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      feature_name: str,
      output_size: int,
      embedding_module: typing.EmbeddingModule,
      features_transform_module: TransformModule = transforms.IdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    # output shapes are arrays to be pytree leaves for tree_map
    output_shapes = {
        f'{feature_name}_{i}': np.asarray(coords.nodal_shape)
        for i in range(output_size)
    }
    self.embedding_fn = embedding_module(
        coords, dt, physics_specs, aux_features, output_shapes=output_shapes
    )
    self.features_transform_fn = features_transform_module(
        coords, dt, physics_specs, aux_features
    )

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> typing.Pytree:
    features = self.embedding_fn(
        inputs, memory, diagnostics, randomness, forcing
    )
    return self.features_transform_fn(features)


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class FloatDataFeatures(hk.Module):
  """Feature module that supplies floating point covariates from data."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      covariate_data_path: str = gin.REQUIRED,
      covariate_keys: Sequence[str] = gin.REQUIRED,
      renaming_dict: Optional[Mapping[str, str]] = None,
      compute_gradients_module: TransformModule = transforms.EmptyTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.covariates = {}
    self.compute_gradients_fn = compute_gradients_module(
        coords, dt, physics_specs, aux_features
    )
    self.coords = coords
    ds = xarray_utils.ds_from_path_or_aux(covariate_data_path, aux_features)
    if renaming_dict is not None:
      ds = ds.rename(renaming_dict)
    lon, lat = (ds[xarray_utils.XR_LON_NAME], ds[xarray_utils.XR_LAT_NAME])
    xarray_utils.verify_grid_consistency(lon, lat, coords.horizontal)
    lon_lat_order = (xarray_utils.XR_LON_NAME, xarray_utils.XR_LAT_NAME)
    for key in covariate_keys:
      data = ds[key].transpose(*lon_lat_order)
      data_units = scales.parse_units(data.attrs['units'])
      data = physics_specs.nondimensionalize(data.values * data_units)
      if data.ndim != 3:
        data = data[np.newaxis, ...]
      self.covariates[key] = data

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> dict[str, jnp.ndarray]:
    del inputs, memory, diagnostics, forcing, randomness  # unused.
    features = {k: v for k, v in self.covariates.items()}
    modal_features = self.coords.horizontal.to_modal(features)
    modal_features = {  # jit should eliminate to_modal if it is not used.
        KeyWithCosLatFactor(k, 0): v for k, v in modal_features.items()
    }
    modal_gradient_features = self.compute_gradients_fn(modal_features)
    sec_lat = 1 / self.coords.horizontal.cos_lat
    sec2_lat = self.coords.horizontal.sec2_lat
    sec_lat_scales = {0: 1, 1: sec_lat, 2: sec2_lat}
    for k, v in modal_gradient_features.items():
      sec_lat_scale = sec_lat_scales[k.factor_order]
      features[k.name] = self.coords.horizontal.to_nodal(v) * sec_lat_scale
    return features


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class CombinedFeatures(hk.Module):
  """Feature module that combines multiple feature modules together."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      feature_modules: Sequence[FeaturesModule] = gin.REQUIRED,
      feature_module_names_to_exclude: Sequence[str] = tuple(),
      features_to_exclude: Sequence[str] = tuple(),
      features_transform_module: TransformModule = transforms.IdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.feature_fns = [
        module(coords, dt, physics_specs, aux_features)
        for module in feature_modules
    ]
    self.feature_module_names_to_exclude = feature_module_names_to_exclude
    self.features_to_exclude = features_to_exclude
    self.features_transform_fn = features_transform_module(
        coords, dt, physics_specs, aux_features
    )

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Forcing] = None,
  ) -> dict[str, jnp.ndarray]:
    all_features = {}
    for feature_fn in self.feature_fns:
      if type(feature_fn).__name__ not in self.feature_module_names_to_exclude:
        features = feature_fn(inputs, memory, diagnostics, randomness, forcing)
        for k, v in features.items():
          if k in all_features:
            raise ValueError(f'Encountered duplicate feature {k}')
          all_features[k] = v
    all_features = self.features_transform_fn(all_features)
    for k in self.features_to_exclude:
      all_features.pop(k, None)
    return all_features


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class NullFeatures(hk.Module):
  """Placeholder features module that returns an empty dict."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      name: Optional[str] = None,
  ):
    del coords, dt, physics_specs, aux_features  # unused
    super().__init__(name=name)

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> dict[str, jnp.ndarray]:
    del inputs, memory, diagnostics, randomness, forcing  # unused
    return {}


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class PressureFeatures(hk.Module):
  """Feature module that computes pressure."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      features_transform_module: TransformModule = transforms.IdentityTransform,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.coords = coords
    self.features_transform_fn = features_transform_module(
        coords, dt, physics_specs, aux_features
    )

  def _nodal_pressure(
      self,
      inputs: typing.Pytree,
      prefix: str = '',
  ) -> Mapping[str, Array]:
    """Computes nodal pressure from model inputs."""
    # Compute nodal, dimensionalized quantities
    to_nodal_fn = self.coords.horizontal.to_nodal
    sigma = self.coords.vertical.centers
    surface_pressure = jnp.exp(to_nodal_fn(inputs['log_surface_pressure']))
    pressure = surface_pressure * sigma[:, jnp.newaxis, jnp.newaxis]
    nodal_features = {prefix + 'pressure': pressure}
    return nodal_features

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> typing.Pytree:
    del memory, diagnostics, randomness, forcing  # unused.
    nodal_features = self._nodal_pressure(inputs)
    return self.features_transform_fn(nodal_features)


@gin.register(denylist=['coords', 'dt', 'physics_specs', 'aux_features'])
class MemoryPressureFeatures(PressureFeatures):
  """Feature module that computes pressure from memory values."""

  def __call__(
      self,
      inputs: typing.Pytree,
      memory: Optional[typing.PyTreeState] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.PyTreeState] = None,
      forcing: Optional[typing.Pytree] = None,
  ) -> typing.Pytree:
    del inputs, diagnostics, randomness, forcing  # unused.
    nodal_features = self._nodal_pressure(memory, 'memory_')
    return self.features_transform_fn(nodal_features)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Array with optional named axes, inspired by Penzai's NamedArray.

This module is intended to be nearly a drop-in replacement for Penzai's
NamedArray, but with an alternative, simpler implementation. Dimensions are
specified by a tuple, where each element is either a string or `None`. `None` is
used to indicate strictly positional dimensions.

Some code and documentation is adapted from penzai.core.named_axes.
"""
from __future__ import annotations

import functools
import operator
import textwrap
import types
from typing import Any, Callable, Self

import jax
import jax.numpy as jnp


def _collect_named_shape(
    leaves_and_paths: list[tuple[jax.tree_util.KeyPath, Any]],
    source_description: str,
) -> dict[str, int]:
  """Collect shared named_shape, or raise an informative error."""
  known_sizes = {}
  bad_dims = []
  for _, leaf in leaves_and_paths:
    if isinstance(leaf, NamedArray):
      for name, size in leaf.named_shape.items():
        if name in known_sizes:
          if known_sizes[name] != size and name not in bad_dims:
            bad_dims.append(name)
        else:
          known_sizes[name] = size

  if bad_dims:
    shapes_str = []
    for keypath, leaf in leaves_and_paths:
      if isinstance(leaf, NamedArray):
        if keypath[0] == jax.tree_util.SequenceKey(0):
          prefix = 'args'
        else:
          assert keypath[0] == jax.tree_util.SequenceKey(1)
          prefix = 'kwargs'
        path = jax.tree_util.keystr(keypath[1:])
        shapes_str.append(f'  {prefix}{path}.named_shape == {leaf.named_shape}')
    shapes_message = '\n'.join(shapes_str)

    raise ValueError(
        f'Inconsistent sizes in a call to {source_description} for dimensions '
        f'{bad_dims}:\n{shapes_message}'
    )

  return known_sizes


def _normalize_out_axes(
    out_axes: dict[str, int] | None, named_shape: dict[str, int]
) -> dict[str, int]:
  """Normalize the out_axes argument to nmap."""
  if out_axes is None:
    return {dim: -(i + 1) for i, dim in enumerate(reversed(named_shape.keys()))}

  if out_axes.keys() != named_shape.keys():
    raise ValueError(
        f'out_axes keys {list(out_axes)} must must match the named '
        f'dimensions {list(named_shape)}'
    )
  any_negative = any(axis < 0 for axis in out_axes.values())
  any_non_negative = any(axis >= 0 for axis in out_axes.values())
  if any_negative and any_non_negative:
    # TODO(shoyer) consider supporting mixed positive and negative out_axes.
    # This would require using jax.eval_shape() to determine the
    # dimensionality of all output arrays.
    raise ValueError(
        'out_axes must be either all positive or all negative, but got '
        f'{out_axes}'
    )
  if len(set(out_axes.values())) != len(out_axes):
    raise ValueError(
        f'out_axes must all have unique values, but got {out_axes}'
    )
  return out_axes


def _nest_vmap_axis(inner_axis: int, outer_axis: int) -> int:
  """Update a vmap in/out axis to account for wrapping in an outer vmap."""
  if outer_axis >= 0 and inner_axis >= 0:
    if inner_axis > outer_axis:
      return inner_axis - 1
    else:
      assert inner_axis < outer_axis
      return inner_axis
  else:
    assert outer_axis < 0 and inner_axis < 0
    if inner_axis > outer_axis:
      return inner_axis
    else:
      assert inner_axis < outer_axis
      return inner_axis + 1


def nmap(fun: Callable, out_axes: dict[str, int] | None = None) -> Callable:  # pylint: disable=g-bare-generic
  """Automatically vectorizes ``fun`` over named dimensions.

  ``nmap`` is a "named dimension vectorizing map". It wraps an ordinary
  positional-axis-based function so that it accepts NamedArrays as input and
  produces NamedArrays as output, and vectorizes over all of named dimensions,
  calling the original function with positionally-indexed slices corresponding
  to each argument's `positional_shape`.

  Unlike `jax.vmap`, the axes to vectorize over are inferred
  automatically from the named dimensions in the NamedArray inputs, rather
  than being specified as part of the mapping transformation. Specifically, each
  dimension name that appears in any of the arguments is vectorized over jointly
  across all arguments that include that dimension, and is then included as a
  named dimension in the output. To make an axis visible to ``fun``, you can
  call `untag` on the argument and pass the axis name(s) of interest; ``fun``
  will then see those axes as positional axes instead of mapping over them.

  `untag` and ``nmap`` are together the primary ways to apply individual
  operations to axes of a NamedArray. `tag` can then be used on the result to
  re-bind names to positional axes.

  Within ``fun``, any mapped-over axes will be accessible using standard JAX
  collective operations like ``psum``, although doing this is usually
  unnecessary.

  Args:
    fun: Function to vectorize by name. This can take arbitrary arguments (even
      non-JAX-arraylike arguments or "static" axis sizes), but must produce a
      PyTree of JAX ArrayLike outputs.
    out_axes: Optional dictionary from dimension names to axis positions in the
      output. By default, dimension names appear as the trailing dimensions of
      every output, in order of their appearance on the inputs.

  Returns:
    An automatically-vectorized version of ``fun``, which can optionally be
    called with NamedArrays instead of ordinary arrays, and which will always
    return NamedArrays for each of its output leaves. Any argument (or PyTree
    leaf of an argument) that is a NamedArray will have its named dimensions
    vectorized over; ``fun`` will then be called with batch tracers
    corresponding to slices of the input array that are shaped like
    ``named_array_arg.positional_shape``.
  """
  if hasattr(fun, '__module__') and hasattr(fun, '__name__'):
    fun_name = f'{fun.__module__}.{fun.__name__}'
  else:
    fun_name = repr(fun)
  if hasattr(fun, '__doc__'):
    fun_doc = fun.__doc__
  else:
    fun_doc = None
  return _nmap_with_doc(fun, fun_name, fun_doc, out_axes)


def _nmap_with_doc(
    fun: Callable,  # pylint: disable=g-bare-generic
    fun_name: str,
    fun_doc: str | None = None,
    out_axes: dict[str, int] | None = None,
) -> Callable:  # pylint: disable=g-bare-generic
  """Implementation of nmap."""

  @functools.wraps(fun)
  def wrapped_fun(*args, **kwargs):
    leaves_and_paths, treedef = jax.tree_util.tree_flatten_with_path(
        (args, kwargs),
        is_leaf=lambda node: isinstance(node, NamedArray),
    )
    leaves = [leaf for _, leaf in leaves_and_paths]

    named_shape = _collect_named_shape(
        leaves_and_paths, source_description=f'nmap({fun})'
    )
    all_dims = tuple(named_shape.keys())
    out_axes_dict = _normalize_out_axes(out_axes, named_shape)

    nested_in_axes = {}
    nested_out_axes = {}

    # working_input_dims and working_out_axes will be iteratively updated to
    # account for nesting in outer vmap calls
    working_input_dims: list[list[str | None]] = [
        list(leaf.dims) if isinstance(leaf, NamedArray) else []
        for leaf in leaves
    ]
    working_out_axes = out_axes_dict.copy()

    # Calculate in_axes and out_axes for all calls to vmap.
    for vmap_dim in all_dims:
      in_axes = []
      for dims in working_input_dims:
        if vmap_dim in dims:
          axis = dims.index(vmap_dim)
          del dims[axis]
        else:
          axis = None
        in_axes.append(axis)
      nested_in_axes[vmap_dim] = in_axes

      out_axis = working_out_axes.pop(vmap_dim)
      for dim in working_out_axes:
        working_out_axes[dim] = _nest_vmap_axis(working_out_axes[dim], out_axis)
      nested_out_axes[vmap_dim] = out_axis

    assert not working_out_axes  # all dimensions processed

    def vectorized_fun(leaf_data):
      args, kwargs = jax.tree.unflatten(treedef, leaf_data)
      return fun(*args, **kwargs)

    # Recursively apply vmap, in the reverse of the order in which we calculated
    # nested_in_axes and nested_out_axes.
    for vmap_dim in reversed(all_dims):
      vectorized_fun = jax.vmap(
          vectorized_fun,
          in_axes=(nested_in_axes[vmap_dim],),
          out_axes=nested_out_axes[vmap_dim],
          axis_name=vmap_dim,
      )

    leaf_data = [
        leaf.data if isinstance(leaf, NamedArray) else leaf for leaf in leaves
    ]
    result = vectorized_fun(leaf_data)

    def wrap_output(data: jnp.ndarray) -> NamedArray:
      dims = [None] * data.ndim
      for dim, axis in out_axes_dict.items():
        dims[axis] = dim
      return NamedArray(data, tuple(dims))

    return jax.tree.map(wrap_output, result)

  docstr = (
      f'Dimension-vectorized version of `{fun_name}`. Takes similar arguments'
      f' as `{fun_name}` but accepts and returns NamedArray objects in place '
      'of arrays.'
  )
  if fun_doc:
    docstr += f'\n\nOriginal documentation:\n\n{fun_doc}'
  wrapped_fun.__doc__ = docstr
  return wrapped_fun


def _swapped_binop(binop):
  """Swaps the order of operations for a binary operation."""

  def swapped(x, y):
    return binop(y, x)

  return swapped


def _wrap_scalar_conversion(scalar_conversion):
  """Wraps a scalar conversion operator on a Field."""

  def wrapped_scalar_conversion(self: NamedArray):
    if self.shape:
      raise ValueError(
          f'Cannot convert a non-scalar NamedArray with {scalar_conversion}'
      )
    return scalar_conversion(self.data)

  return wrapped_scalar_conversion


def _wrap_array_method(name):
  """Wraps an array method on a Field."""

  def func(array, *args, **kwargs):
    return getattr(array, name)(*args, **kwargs)

  array_method = getattr(jax.Array, name)
  wrapped_func = nmap(func)
  functools.update_wrapper(
      wrapped_func,
      array_method,
      assigned=('__name__', '__qualname__', '__annotations__'),
      updated=(),
  )
  wrapped_func.__module__ = __name__
  wrapped_func.__doc__ = (
      'Name-vectorized version of array method'
      f' `{name} <numpy.ndarray.{name}>`. Takes similar arguments as'
      f' `{name} <numpy.ndarray.{name}>` but accepts and returns NamedArray'
      ' objects in place of regular arrays.'
  )
  return wrapped_func


_VALID_PYTREE_OPS = (
    'JAX pytree operations on NamedArray objects are only valid when they'
    ' insert new leading dimensions, or trim unnamed leading dimensions. The'
    ' sizes and positions (from the end) of all named dimensions must be'
    ' preserved.'
)


@jax.tree_util.register_pytree_node_class
class NamedArray:
  """Array with optionally named axes.

  Axis names are either a string or None, indicating an unnamed axis.

  Attributes:
    data: the underlying data array.
    dims: tuple of dimension names, with the same length as `data.ndim`. Strings
      indicate named axes, and may not be repeated. `None` indicates positional
      axes.
  """

  _data: jnp.ndarray
  _dims: tuple[str | None, ...]

  def __init__(
      self,
      data: jax.typing.ArrayLike,
      dims: tuple[str | None, ...] | None = None,
  ):
    """Construct a NamedArray.

    Arguments:
      data: the underlying data array.
      dims: optional tuple of dimension names, with the same length as
        `data.ndim`. Strings indicate named axes, and may not be repeated.
        `None` indicates positional axes. If `dims` is not provided, all axes
        are positional.
    """
    data = jnp.asarray(data)
    if dims is None:
      dims = (None,) * data.ndim
    else:
      if data.ndim != len(dims):
        raise ValueError(f'{data.ndim=} != {len(dims)=}')
      named_dims = [dim for dim in dims if dim is not None]
      if len(set(named_dims)) < len(named_dims):
        raise ValueError('dimension names may not be repeated: {dims}')
    self._data = data
    self._dims = dims

  @property
  def data(self) -> jnp.ndarray:
    """Data associated with this array."""
    return self._data

  @property
  def dims(self) -> tuple[str | None, ...]:
    """Dimension names of this array."""
    return self._dims

  @property
  def ndim(self) -> int:
    """Number of dimensions in the array, including postional and named axes."""
    return self.data.ndim

  @property
  def shape(self) -> tuple[int, ...]:
    """Shape of the array, including positional and named axes."""
    return self.data.shape

  @property
  def positional_shape(self) -> tuple[int, ...]:
    """Shape of the array with all named axes removed."""
    return tuple(
        size for dim, size in zip(self.dims, self.data.shape) if dim is None
    )

  @property
  def named_shape(self) -> dict[str, int]:
    """Mapping from dimension names to sizes."""
    return {
        dim: size
        for dim, size in zip(self.dims, self.data.shape)
        if dim is not None
    }

  def __repr__(self) -> str:
    indent = lambda x: textwrap.indent(x, prefix=' ' * 13)[13:]
    return textwrap.dedent(f"""\
    {type(self).__name__}(
        data={indent(repr(self.data))},
        dims={self.dims},
    )""")

  def tree_flatten(self):
    """Flatten this object for JAX pytree operations."""
    if isinstance(self.data, jnp.ndarray):
      size_info = tuple(self.named_shape.items())
    else:
      # Arrays unflattened from non-ndarray leaves are flattened inside vmap.
      # The resulting treedef is ignored.
      size_info = None
    return [self.data], (self.dims, size_info)

  @classmethod
  def _new_unvalidated(
      cls, data: jnp.ndarray, dims: tuple[str | None, ...]
  ) -> Self:
    """Create a new NamedArray, without validating data and dims."""
    # pylint: disable=protected-access
    obj = super().__new__(cls)
    obj._data = data
    obj._dims = dims
    return obj

  @classmethod
  def _new_with_padded_or_trimmed_dims(
      cls, data: jnp.ndarray, dims: tuple[str | None, ...]
  ) -> Self:
    """Create a new NamedArray, padding or trimming dims to match data.ndim."""
    assert isinstance(data, jnp.ndarray)
    if len(dims) <= data.ndim:
      dims = (None,) * (data.ndim - len(dims)) + dims
    else:
      trimmed_dims = dims[: -data.ndim]
      if any(dim is not None for dim in trimmed_dims):
        raise ValueError(
            'cannot trim named dimensions when unflattening to a NamedArray:'
            f' {trimmed_dims}. {_VALID_PYTREE_OPS} If you are using vmap or'
            ' scan, the first dimension must be unnamed.'
        )
      dims = dims[-data.ndim :]
    return cls(data, dims)

  @classmethod
  def tree_unflatten(cls, treedef, leaves: list[jax.Array | object]) -> Self:
    """Unflatten this object for JAX pytree operations."""
    dims, size_info = treedef
    [data] = leaves

    if not isinstance(data, jnp.ndarray):
      # JAX builds pytrees with non-ndarray leaves inside some transformations,
      # such as vmap, for handling the in_axes argument.
      return cls._new_unvalidated(data, dims)

    # Restored NamedArray objects may have additional or removed leading
    # dimensions, if produced with scan or vmap.
    result = cls._new_with_padded_or_trimmed_dims(data, dims)
    assert size_info is not None
    named_shape = dict(size_info)
    if result.named_shape != named_shape:
      raise ValueError(
          'named shape mismatch when unflattening to a NamedArray: '
          f'{result.named_shape} != {named_shape}. {_VALID_PYTREE_OPS}'
      )
    return result

  def tag(self, *dims: str) -> Self:
    """Attaches dimension names to the positional axes of an array.

    Args:
      *dims: axis names to assign to each positional axis in the array. Must
        have exactly the same length as the number of unnamed axes in the array.

    Raises:
      ValueError: If the wrong number of dimensions are provided.

    Returns:
      A NamedArray with the given names assigned to the positional axes, and no
      remaining positional axes.
    """
    if len(dims) != len(self.positional_shape):
      pos_ndim = len(self.positional_shape)
      raise ValueError(
          'there must be exactly as many dimensions given to `tag` as there'
          f' are positional axes in the array, but got {dims} for '
          f'{pos_ndim} positional {"axis" if pos_ndim == 1 else "axes"}.'
      )

    if any(not isinstance(name, str) for name in dims):
      raise TypeError(f'dimension names must be strings: {dims}')

    dim_queue = list(reversed(dims))
    new_dims = tuple(
        dim_queue.pop() if dim is None else dim for dim in self.dims
    )
    assert not dim_queue
    return type(self)(self.data, new_dims)

  def untag(self, *dims: str) -> Self:
    """Removes the requested dimension names.

    `untag` can only be called on a `NamedArray` that does not have any
    positional axes. It produces a new `NamedArray` where the axes with the
    requested dimension names are now treated as positional instead.

    Args:
      *dims: axis names to make positional, in the order they should appear in
        the positional array.

    Raises:
      ValueError: if the provided axis ordering is not valid.

    Returns:
      A named array with the given dimensions converted to positional axes.
    """
    if self.positional_shape:
      raise ValueError(
          '`untag` cannot be used to introduce positional axes for a NamedArray'
          ' that already has positional axes. Please assign names to the'
          ' existing positional axes first using `tag`.'
      )

    named_shape = self.named_shape
    if any(dim not in named_shape for dim in dims):
      raise ValueError(
          f'cannot untag {dims} because they are not a subset of the current '
          f'named dimensions {tuple(self.dims)}'
      )

    ordered = tuple(sorted(dims, key=self.dims.index))
    if ordered != dims:
      raise ValueError(
          f'cannot untag {dims} because they do not appear in the order of '
          f'the current named dimensions {ordered}'
      )

    untagged = set(dims)
    new_dims = tuple(None if dim in untagged else dim for dim in self.dims)
    return type(self)(self.data, new_dims)

  def order_as(self, *dims: str | types.EllipsisType) -> Self:
    """Reorder the dimensions of an array.

    All dimensions must be named. Use `tag` first to name any positional axes.

    Args:
      *dims: dimension names that appear on this array, in the desired order on
        the result. `...` may be used once, to indicate all other dimensions in
        order of appearance on this array.

    Returns:
      Array with transposed data and reordered dimensions, as indicated.
    """
    if any(dim is None for dim in self.dims):
      raise ValueError(
          'cannot reorder the dimensions of an array with unnamed '
          f'dimensions: {self.dims}'
      )

    ellipsis_count = sum(dim is ... for dim in dims)
    if ellipsis_count > 1:
      raise ValueError(
          f'dimension names contain multiple ellipses (...): {dims}'
      )
    elif ellipsis_count == 1:
      explicit_dims = {dim for dim in dims if dim is not ...}
      implicit_dims = tuple(
          dim for dim in self.dims if dim not in explicit_dims
      )
      i = dims.index(...)
      dims = dims[:i] + implicit_dims + dims[i + 1 :]

    order = tuple(self.dims.index(dim) for dim in dims)
    return type(self)(self.data.transpose(order), dims)

  # Convenience wrappers: Elementwise infix operators.
  __lt__ = _nmap_with_doc(operator.lt, 'jax.Array.__lt__')
  __le__ = _nmap_with_doc(operator.le, 'jax.Array.__le__')
  __eq__ = _nmap_with_doc(operator.eq, 'jax.Array.__eq__')
  __ne__ = _nmap_with_doc(operator.ne, 'jax.Array.__ne__')
  __ge__ = _nmap_with_doc(operator.ge, 'jax.Array.__ge__')
  __gt__ = _nmap_with_doc(operator.gt, 'jax.Array.__gt__')

  __add__ = _nmap_with_doc(operator.add, 'jax.Array.__add__')
  __sub__ = _nmap_with_doc(operator.sub, 'jax.Array.__sub__')
  __mul__ = _nmap_with_doc(operator.mul, 'jax.Array.__mul__')
  __truediv__ = _nmap_with_doc(operator.truediv, 'jax.Array.__truediv__')
  __floordiv__ = _nmap_with_doc(operator.floordiv, 'jax.Array.__floordiv__')
  __mod__ = _nmap_with_doc(operator.mod, 'jax.Array.__mod__')
  __divmod__ = _nmap_with_doc(divmod, 'jax.Array.__divmod__')
  __pow__ = _nmap_with_doc(operator.pow, 'jax.Array.__pow__')
  __lshift__ = _nmap_with_doc(operator.lshift, 'jax.Array.__lshift__')
  __rshift__ = _nmap_with_doc(operator.rshift, 'jax.Array.__rshift__')
  __and__ = _nmap_with_doc(operator.and_, 'jax.Array.__and__')
  __or__ = _nmap_with_doc(operator.or_, 'jax.Array.__or__')
  __xor__ = _nmap_with_doc(operator.xor, 'jax.Array.__xor__')

  __radd__ = _nmap_with_doc(_swapped_binop(operator.add), 'jax.Array.__radd__')
  __rsub__ = _nmap_with_doc(_swapped_binop(operator.sub), 'jax.Array.__rsub__')
  __rmul__ = _nmap_with_doc(_swapped_binop(operator.mul), 'jax.Array.__rmul__')
  __rtruediv__ = _nmap_with_doc(
      _swapped_binop(operator.truediv), 'jax.Array.__rtruediv__'
  )
  __rfloordiv__ = _nmap_with_doc(
      _swapped_binop(operator.floordiv), 'jax.Array.__rfloordiv__'
  )
  __rmod__ = _nmap_with_doc(_swapped_binop(operator.mod), 'jax.Array.__rmod__')
  __rdivmod__ = _nmap_with_doc(_swapped_binop(divmod), 'jax.Array.__rdivmod__')
  __rpow__ = _nmap_with_doc(_swapped_binop(operator.pow), 'jax.Array.__rpow__')
  __rlshift__ = _nmap_with_doc(
      _swapped_binop(operator.lshift), 'jax.Array.__rlshift__'
  )
  __rrshift__ = _nmap_with_doc(
      _swapped_binop(operator.rshift), 'jax.Array.__rrshift__'
  )
  __rand__ = _nmap_with_doc(_swapped_binop(operator.and_), 'jax.Array.__rand__')
  __ror__ = _nmap_with_doc(_swapped_binop(operator.or_), 'jax.Array.__ror__')
  __rxor__ = _nmap_with_doc(_swapped_binop(operator.xor), 'jax.Array.__rxor__')

  __abs__ = _nmap_with_doc(operator.abs, 'jax.Array.__abs__')
  __neg__ = _nmap_with_doc(operator.neg, 'jax.Array.__neg__')
  __pos__ = _nmap_with_doc(operator.pos, 'jax.Array.__pos__')
  __invert__ = _nmap_with_doc(operator.inv, 'jax.Array.__invert__')

  # Convenience wrappers: Scalar conversions.
  __bool__ = _wrap_scalar_conversion(bool)
  __complex__ = _wrap_scalar_conversion(complex)
  __int__ = _wrap_scalar_conversion(int)
  __float__ = _wrap_scalar_conversion(float)
  __index__ = _wrap_scalar_conversion(operator.index)

  # elementwise operations
  astype = _wrap_array_method('astype')
  clip = _wrap_array_method('clip')
  conj = _wrap_array_method('conj')
  conjugate = _wrap_array_method('conjugate')
  imag = _wrap_array_method('imag')
  real = _wrap_array_method('real')
  round = _wrap_array_method('round')
  view = _wrap_array_method('view')



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Modules that define stacks of neural-net layers acting on spatial arrays."""

import functools
import itertools
from typing import Callable, Protocol, Sequence

from flax import nnx
import jax
import jax.numpy as jnp
from neuralgcm.experimental import standard_layers
from neuralgcm.experimental import typing

Array = typing.Array


class UnaryTower(Protocol):
  """Protocol for towers that transform array -> array preserving spatial shape.

  This is a protocol, so we can use any module that implements the same
  signature without complicating the inheretance hierarchy.
  """

  def __init__(self, input_size: int, output_size: int, **kwargs):
    ...

  def __call__(self, inputs: jax.Array) -> jax.Array:
    ...


UnaryTowerFactory = Callable[..., UnaryTower]


# TODO(dkochkov) consider passing spatial axes to parameterize # of vmap needed.


class ColumnTower(nnx.Module):
  """Tower that operates on latent/level axis with shared NN across all lat/lon.

  input shape: [nn_input_shape, lon, lat],
  output shape: [nn_output_shape, lon, lat].
  """

  def __init__(
      self,
      input_size: int,
      output_size: int,
      *,
      column_net_factory: nnx.Module,
      apply_remat: bool = False,
      rngs: nnx.Rngs,
  ):
    self.column_network = column_net_factory(input_size, output_size, rngs=rngs)
    self.apply_remat = apply_remat

  def __call__(self, inputs: Array) -> Array:
    """Applies Column tower to inputs."""
    vmap = functools.partial(nnx.vmap, in_axes=(None, -1), out_axes=-1)

    def vmap_fn(net, inputs):
      return net(inputs)

    mapped_column_net = vmap(vmap(vmap_fn))
    if self.apply_remat:
      mapped_column_net = nnx.remat(mapped_column_net)
    return mapped_column_net(self.column_network, inputs)


class EpdTower(nnx.Module):
  """EPD tower module parameterized by encode/process/decode towers."""

  def __init__(
      self,
      input_size: int,
      output_size: int,
      *,
      latent_size: int,
      num_process_blocks: int,
      encode_tower_factory: UnaryTowerFactory,
      process_tower_factory: UnaryTowerFactory,
      decode_tower_factory: UnaryTowerFactory,
      post_encode_activation: Callable[[Array], Array] | None = None,
      pre_decode_activation: Callable[[Array], Array] | None = None,
      final_activation: Callable[[Array], Array] | None = None,
      rngs: nnx.Rngs,
  ):
    self.input_size = input_size
    self.output_size = output_size
    self.latent_size = latent_size
    self.post_encode_activation = post_encode_activation
    self.pre_decode_activation = pre_decode_activation
    self.final_activation = final_activation
    self.encoder = encode_tower_factory(input_size, latent_size, rngs=rngs)
    self.decoder = decode_tower_factory(latent_size, output_size, rngs=rngs)
    self.process_towers = tuple([
        process_tower_factory(latent_size, latent_size, rngs=rngs)
        for _ in range(num_process_blocks)
    ])

  def __call__(self, inputs: Array) -> Array:
    """Applies EPD tower to inputs."""
    encoded = self.encoder(inputs)
    if self.post_encode_activation is not None:
      encoded = self.post_encode_activation(encoded)
    current = encoded
    for process_block in self.process_towers:
      current = current + process_block(current)
    if self.pre_decode_activation is not None:
      current = self.pre_decode_activation(current)
    out = self.decoder(current)
    if self.final_activation is not None:
      return self.final_activation(out)
    return out


class ChainedTower(nnx.Module):
  """Chains multiple towers together."""

  def __init__(
      self,
      input_size: int,
      output_size: int,
      *,
      intermediate_sizes: tuple[int | str, ...],
      tower_modules: Sequence[UnaryTowerFactory],
      rngs: nnx.Rngs,
  ):
    self.input_size = input_size
    self.output_size = output_size
    replace_values = {'input_size': input_size, 'output_size': output_size}
    replace_str = lambda s: replace_values[s]
    intermediate_sizes = tuple(
        replace_str(s) if isinstance(s, str) else s for s in intermediate_sizes
    )
    output_sizes = (*intermediate_sizes, output_size)
    input_sizes = (input_size, *intermediate_sizes)
    zip_vals = zip(input_sizes, output_sizes, tower_modules, strict=True)
    self.towers = tuple(
        tower_factory(d_in, d_out, rngs=rngs)
        for d_in, d_out, tower_factory in zip_vals
    )

  def __call__(self, inputs: Array) -> Array:
    current = inputs
    for tower in self.towers:
      current = tower(current)
    return current


class ConvLonLatTower(nnx.Module):
  """Two dimensional ConvNet tower module."""

  def __init__(
      self,
      input_size: int,
      output_size: int,
      *,
      num_hidden_units: int,
      num_hidden_layers: int,
      kernel_size: tuple[int, int] = (3, 3),
      dilation: int = 1,
      activation: Callable[[jnp.ndarray], jnp.ndarray] = jax.nn.gelu,
      use_bias: bool = True,
      activate_final: bool = False,
      apply_remat: bool = False,
      rngs: nnx.Rngs,
  ):
    self.apply_remat = apply_remat
    self.activation = activation
    self.activate_final = activate_final
    self.num_hidden_layers = num_hidden_layers
    self.conv_layers = []
    self.conv_layers.append(
        standard_layers.ConvLonLat(
            input_size=input_size,
            output_size=num_hidden_units,
            use_bias=use_bias,
            kernel_size=kernel_size,
            dilation=dilation,
            rngs=rngs,
        )
    )
    for _ in range(num_hidden_layers):
      self.conv_layers.append(
          standard_layers.ConvLonLat(
              input_size=num_hidden_units,
              output_size=num_hidden_units,
              use_bias=use_bias,
              kernel_size=kernel_size,
              dilation=dilation,
              rngs=rngs,
          )
      )
    self.conv_layers.append(
        standard_layers.ConvLonLat(
            input_size=num_hidden_units,
            output_size=output_size,
            use_bias=use_bias,
            kernel_size=kernel_size,
            dilation=dilation,
            rngs=rngs,
        )
    )

  def apply(self, inputs: Array) -> Array:
    for i, layer in enumerate(self.conv_layers):
      inputs = layer(inputs)
      if i != self.num_hidden_layers + 1 or self.activate_final:
        inputs = self.activation(inputs)
    return inputs

  def __call__(self, inputs: Array) -> Array:
    apply_fn = self.apply
    if self.apply_remat:
      apply_fn = nnx.remat(apply_fn)
    return apply_fn(inputs)


class ResNet(nnx.Module):
  """A ResNet tower module that uses ConvLonLatTower for each block.

  input size: number of input channels.
  output size: number of output channels.

  conv_block_num_channels: number of channels for output of the intermediate and
    final layer of conv block components.

  conv_block_num_hidden_layers: number of hidden layers in each conv block.
  kernel size: tuple of convolutional kernel used in all layers.
  dilations: tuple of dilations used if an int is passed all layers will use the
    same dilation.

  activation: activation function throughout the tower and subcomponents.
  use_bias: whether to use bias.
  apply_remat: whether to apply rematerialization during training.
  num_hidden_layers_skip_residual: from the final layer, how many additional
    layers to skip passing the residual connection.
  """

  def __init__(
      self,
      input_size: int,
      output_size: int,
      *,
      conv_block_num_channels: tuple[int, ...],
      conv_block_num_hidden_layers: int = 0,
      kernel_size: tuple[int, int] = (3, 3),
      dilations: int | tuple[int, ...] = 1,
      activation: Callable[[jnp.ndarray], jnp.ndarray] = jax.nn.gelu,
      use_bias: bool = True,
      apply_remat: bool = False,
      num_hidden_layers_skip_residual: int = 0,
      rngs: nnx.Rngs,
  ):

    if isinstance(dilations, int):
      dilations = tuple([dilations] * (len(conv_block_num_channels) + 1))

    in_through_hidden_channels = (input_size,) + conv_block_num_channels
    self.num_hidden_layers_skip_residual = num_hidden_layers_skip_residual
    self.activation = activation
    self.apply_remat = apply_remat

    self.convolutional_blocks = []
    self.projection_blocks = []
    for dilation, (din, dout) in zip(
        dilations[:-1],
        itertools.pairwise(in_through_hidden_channels),
        strict=True,
    ):
      self.convolutional_blocks.append(
          ConvLonLatTower(
              input_size=din,
              output_size=dout,
              num_hidden_units=dout,
              num_hidden_layers=conv_block_num_hidden_layers,
              use_bias=use_bias,
              kernel_size=kernel_size,
              dilation=dilation,
              activation=activation,
              activate_final=False,
              rngs=rngs,
          )
      )
      if din != dout:
        self.projection_blocks.append(
            standard_layers.ConvLonLat(
                input_size=din,
                output_size=dout,
                use_bias=False,
                kernel_size=(1, 1),
                dilation=1,
                rngs=rngs,
            )
        )
      else:
        self.projection_blocks.append(lambda x: x)

    self.final_convolutional_block = ConvLonLatTower(
        input_size=conv_block_num_channels[-1],
        output_size=output_size,
        num_hidden_units=output_size,
        num_hidden_layers=conv_block_num_hidden_layers,
        use_bias=use_bias,
        kernel_size=kernel_size,
        dilation=dilations[-1],
        activation=activation,
        activate_final=False,
        rngs=rngs,
    )

  def apply(self, inputs: Array) -> Array:
    carry = inputs
    for i, (convolution, projection) in enumerate(
        zip(self.convolutional_blocks, self.projection_blocks)
    ):
      residual_connection = projection(carry)
      carry = convolution(carry)
      if i < len(self.projection_blocks) - self.num_hidden_layers_skip_residual:
        carry += residual_connection
      carry = self.activation(carry)

    carry = self.final_convolutional_block(carry)
    return carry

  def __call__(self, inputs: Array) -> Array:
    apply_fn = self.apply
    if self.apply_remat:
      apply_fn = nnx.remat(apply_fn)
    return apply_fn(inputs)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines equation and time integration modules."""

import dataclasses
from typing import Callable

from dinosaur import time_integration
from flax import nnx
from neuralgcm.experimental import typing
import tree_math


def repeated(fn: typing.StepFn, steps: int) -> typing.StepFn:
  """Returns a function that applies `fn` `steps` times."""
  if steps == 1:
    return fn

  return nnx.scan(fn, length=steps, in_axes=nnx.Carry, out_axes=nnx.Carry)


# TODO(dkochkov): Remove this once we migrate to using `Timedelta` everywhere.
# Alias for fixing rounding errors in sim_time.
maybe_fix_sim_time_roundoff = time_integration.maybe_fix_sim_time_roundoff


class ExplicitODE(time_integration.ExplicitODE, nnx.Module):
  """Module wrapper for ExplicitODE.

  This module is wrapped as nnx.Module to ensure that any submodule that is
  a part of the equation class is included in the model's parameter tree.
  """


class ImplicitExplicitODE(time_integration.ImplicitExplicitODE, nnx.Module):
  """Module wrapper for ImplicitExplicitODE.

  This module is wrapped as nnx.Module to ensure that any submodule that is
  a part of the equation class is included in the model's parameter tree.
  """


def forward_euler(equation: ExplicitODE, time_step: float) -> typing.StepFn:
  """Time stepping for an explicit ODE via forward Euler method.

  This method is first order accurate.

  Args:
    equation: equation to solve.
    time_step: time step.

  Returns:
    Function that performs a time step.
  """
  # pylint: disable=invalid-name
  dt = time_step
  F = tree_math.unwrap(equation.explicit_terms)

  @tree_math.wrap
  def step_fn(u0):
    return u0 + dt * F(u0)

  return step_fn


@dataclasses.dataclass
class DinosaurIntegrator(nnx.Module):
  """Module that wraps time integrators from dinosaur package."""

  equation: ExplicitODE | ImplicitExplicitODE
  time_step: float
  integrator: Callable[
      [ExplicitODE | ImplicitExplicitODE, float], typing.StepFn
  ]

  def __call__(self, inputs: typing.Pytree) -> typing.Pytree:
    return self.integrator(self.equation, self.time_step)(inputs)


# Note: we don't use functools.partial here because it would cause issues with
# fiddle serialization.
class ImexRk3Sil(DinosaurIntegrator):

  def __init__(
      self,
      equation: ExplicitODE | ImplicitExplicitODE,
      time_step: float,
  ):
    super().__init__(
        equation=equation,
        time_step=time_step,
        integrator=time_integration.imex_rk_sil3,
    )


class ExplicitEuler(DinosaurIntegrator):

  def __init__(
      self,
      equation: ExplicitODE,
      time_step: float,
  ):
    super().__init__(
        equation=equation,
        time_step=time_step,
        integrator=forward_euler,
    )



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the neuralgcm module."""

# pylint: disable=unused-import
# pylint: disable=g-bad-import-order

import neuralgcm.api
import neuralgcm.correctors
import neuralgcm.decoders
import neuralgcm.demo
import neuralgcm.diagnostics
import neuralgcm.embeddings
import neuralgcm.encoders
import neuralgcm.equations
import neuralgcm.features
import neuralgcm.filters
import neuralgcm.forcings
import neuralgcm.gin_utils
import neuralgcm.initializers
import neuralgcm.integrators
import neuralgcm.layers
import neuralgcm.mappings
import neuralgcm.model_builder
import neuralgcm.model_utils
import neuralgcm.optimization
import neuralgcm.orographies
import neuralgcm.parameterizations
import neuralgcm.perturbations
import neuralgcm.stochastic
import neuralgcm.towers

from neuralgcm.api import PressureLevelModel

__version__ = "1.1.1"  # keep in sync with pyproject.toml



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Modules that predict refinement or updates of time-advanced states."""

import dataclasses
from typing import Any, Callable, Optional
from dinosaur import coordinate_systems
from dinosaur import time_integration
from dinosaur import typing
import gin
import haiku as hk
import jax
from neuralgcm import equations
from neuralgcm import features
from neuralgcm import filters
from neuralgcm import integrators
from neuralgcm import mappings

Pytree = typing.Pytree
PyTreeState = typing.PyTreeState
Forcing = typing.Forcing

CorrectorFn = typing.CorrectorFn
CorrectorModule = typing.CorrectorModule
EquationModule = equations.EquationModule
FeaturesModule = features.FeaturesModule
MappingModule = mappings.MappingModule
StepModule = typing.StepModule
StepFilterModule = Callable[..., typing.PyTreeStepFilterFn]
TimeIntegrator = integrators.TimeIntegrator
TransformModule = typing.TransformModule


@gin.register
class PredictorEulerCorrector(hk.Module):
  """Corrector that takes Euler step ontop of a predictor step."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      predictor_module: StepModule,
      filter_module: StepFilterModule = filters.NoFilter,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.dt = dt
    self.step_fn = predictor_module(coords, dt, physics_specs, aux_features)
    self.filter_fn = filter_module(coords, dt, physics_specs, aux_features)

  def __call__(
      self,
      state: typing.PyTreeState,
      tendencies: typing.PyTreeState,
      forcing: Optional[Forcing] = None,
  ) -> typing.PyTreeState:
    state = self.step_fn(state, forcing)
    euler_add_fn = lambda x, y: x + self.dt * y if y is not None else x
    result = jax.tree_util.tree_map(euler_add_fn, state, tendencies)
    return self.filter_fn(state, result)


@gin.register
class DycoreWithPhysicsCorrector(hk.Module):
  """Corrector that runs dycore with physics tendencies added to explicit terms.

  This corrector treats predicted physics tendencies constant at each time
  interval and includes them to all substeps of the dycore step. To achieve this
  the dycore in this module is specified by the governing equation, rather than
  an `EquationStep`.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      dycore_equation_module: EquationModule = gin.REQUIRED,
      dycore_substeps: int = gin.REQUIRED,
      time_integrator: TimeIntegrator = integrators.imex_rk_sil3,
      filter_module: StepFilterModule = filters.NoFilter,
      checkpoint_explicit_terms: bool = True,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    dycore_equation = dycore_equation_module(
        coords, dt, physics_specs, aux_features)
    if checkpoint_explicit_terms:
      dycore_equation = time_integration.ImplicitExplicitODE.from_functions(
          hk.remat(dycore_equation.explicit_terms),
          dycore_equation.implicit_terms,
          dycore_equation.implicit_inverse)
    self.coords = coords
    self.dycore_equation = dycore_equation
    self.dycore_substeps = dycore_substeps
    self.inner_dt = dt / dycore_substeps
    self.dt = dt
    self.time_integrator = time_integrator
    self.filter_fn = filter_module(coords, dt, physics_specs, aux_features)

  def __call__(
      self,
      state: typing.PyTreeState,
      tendencies: typing.PyTreeState,
      forcing: Optional[Forcing] = None,
  ) -> typing.PyTreeState:
    state, tendencies = self.coords.with_dycore_sharding((state, tendencies))
    physics_parametrization_eq = time_integration.ExplicitODE.from_functions(
        lambda state: tendencies)
    all_equations = (self.dycore_equation, physics_parametrization_eq)
    equation = time_integration.compose_equations(all_equations)
    step_fn = self.time_integrator(equation, self.inner_dt)
    # TODO(dkochkov) make step_with_filters work with single filter.
    step_fn = time_integration.step_with_filters(step_fn, [self.filter_fn])
    step_fn = time_integration.repeated(step_fn, self.dycore_substeps, hk.scan)
    state = time_integration.maybe_fix_sim_time_roundoff(
        step_fn(state), self.dt
    )
    state = self.coords.with_dycore_sharding(state)
    return state


@gin.register
class CustomCoordsCorrector(hk.Module):
  """Corrector module that uses gin-configured coordinates instead of coords.

  This class currently supports model states in spectral representation. It
  could be easily extended to nodal-state models by converting to modal space
  prior to spectral interpolation and back after the timestep if performed.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      corrector_module: CorrectorModule = gin.REQUIRED,
      custom_coords: coordinate_systems.CoordinateSystem = gin.REQUIRED,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    custom_coords = dataclasses.replace(
        custom_coords, spmd_mesh=coords.spmd_mesh
    )
    self.corrector_fn = corrector_module(
        custom_coords, dt, physics_specs, aux_features)
    self.to_custom_coords_fn = coordinate_systems.get_spectral_interpolate_fn(
        coords, custom_coords)
    self.from_custom_coords_fn = coordinate_systems.get_spectral_interpolate_fn(
        custom_coords, coords)

  def __call__(
      self,
      state: typing.PyTreeState,
      tendencies: typing.PyTreeState,
      forcing: Optional[Forcing] = None,
  ) -> typing.PyTreeState:
    state = self.to_custom_coords_fn(state)
    tendencies = self.to_custom_coords_fn(tendencies)
    # TODO(dkochkov) Consider adding forcing interpolated to custom coords.
    custom_out = self.corrector_fn(state, tendencies, None)
    return self.from_custom_coords_fn(custom_out)
